#! /usr/bin/env python
"""SearchSource: Springer Link"""
from __future__ import annotations

import re
import typing
import urllib.parse
from pathlib import Path

import inquirer
import pandas as pd
import requests
import zope.interface
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
import colrev.settings
from colrev.constants import Colors
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType

# pylint: disable=unused-argument
# pylint: disable=duplicate-code

# Note : API requires registration
# https://dev.springernature.com/


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class SpringerLinkSearchSource:
    """Springer Link"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings

    endpoint = "colrev.springer_link"
    # pylint: disable=colrev-missed-constant-usage
    source_identifier = "url"
    search_types = [SearchType.DB, SearchType.API]

    ci_supported: bool = Field(default=False)
    heuristic_status = SearchSourceHeuristicStatus.supported

    SETTINGS = {
        "api_key": "packages.search_source.colrev.springer_link.api_key",
    }
    db_url = "https://link.springer.com/"

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.review_manager = source_operation.review_manager
        self.search_source = self.settings_class(**settings)
        self.quality_model = self.review_manager.get_qm()
        self.source_operation = source_operation
        self.language_service = colrev.env.language_service.LanguageService()

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Springer Link"""

        result = {"confidence": 0.0}

        if filename.suffix == ".csv":
            if data.count("http://link.springer.com") > data.count("\n") - 2:
                result["confidence"] = 1.0
                return result

        # Note : no features in bib file for identification

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        params_dict: dict[str, str] = {}
        search_type = operation.select_search_type(
            search_types=cls.search_types, params=params_dict
        )

        if search_type == SearchType.DB:
            search_source = operation.create_db_source(
                search_source_cls=cls,
                params={},
            )

        elif search_type == SearchType.API:
            filename = operation.get_unique_filename(file_path_string="springer_link")
            search_source = colrev.settings.SearchSource(
                endpoint=cls.endpoint,
                filename=filename,
                search_type=SearchType.API,
                search_parameters={},
                comment="",
            )
            params_dict.update(vars(search_source))
            instance = cls(source_operation=operation, settings=params_dict)
            instance.api_ui()
            search_source.search_parameters = instance._add_constraints()

        else:
            raise NotImplementedError

        operation.add_source_and_search(search_source)
        return search_source

    def search(self, rerun: bool) -> None:
        """Run a search of SpringerLink"""

        if self.search_source.search_type == SearchType.DB:
            self.source_operation.run_db_search(  # type: ignore
                search_source_cls=self.__class__,
                source=self.search_source,
            )
            return

        if self.search_source.search_type == SearchType.API:
            springer_feed = self.search_source.get_api_feed(
                review_manager=self.review_manager,
                source_identifier=self.source_identifier,
                update_only=(not rerun),
            )
            self._run_api_search(springer_feed=springer_feed, rerun=rerun)
            return

        raise NotImplementedError

    def _add_constraints(self) -> dict:
        """add constraints for API query"""
        complex_query_prompt = [
            inquirer.List(
                name="complex_query",
                message="Select how the search will be entered",
                choices=["interactively", "complete_search_string"],
            ),
        ]

        answers = inquirer.prompt(complex_query_prompt)

        if answers["complex_query"] == "complete_search_string":
            query = input("Please enter your search string: ")
            search_parameters = {"query": query}

        else:

            print(
                "Please enter your search parameter for the following constraints"
                + "(or just press enter to continue):"
            )
            keyword = input("keyword: ")
            subject_choices = [
                inquirer.Checkbox(
                    name="subject",
                    message="Select subject(s) or press Enter to skip:",
                    choices=[
                        "Astronomy",
                        "Behavioral Sciences",
                        "Biomedical Sciences",
                        "Business and Management",
                        "Chemistry",
                        "Climate",
                        "Computer Science",
                        "Earth Sciences",
                        "Economics",
                        "Education and Language",
                        "Energy",
                        "Engineering",
                        "Environmental Sciences",
                        "Food Science and Nutrition",
                        "General Interest",
                        "Geography",
                        "Law",
                        "Life Sciences",
                        "Materials",
                        "Mathematics",
                        "Medicine",
                        "Philosophy",
                        "Physics",
                        "Public Health",
                        "Social Sciences",
                        "Statistics",
                        "Water",
                    ],
                ),
            ]
            answers = inquirer.prompt(subject_choices)
            subject = ",".join(answers["subject"])

            language_choices = [
                inquirer.Checkbox(
                    name="language",
                    message="Select language(s) or press Enter to skip:",
                    choices=[
                        "en",
                        "de",
                        "fr",
                        "es",
                        "it",
                        "pt",
                        "nl",
                    ],
                ),
            ]
            answers = inquirer.prompt(language_choices)
            language = ",".join(answers["language"])

            year_choices = [
                inquirer.Text(
                    name="year",
                    message="Select year(s) or press Enter to skip:",
                    validate=self._is_year,
                ),
            ]
            answers = inquirer.prompt(year_choices)
            year = answers["year"]

            doc_type_choices = [
                inquirer.Checkbox(
                    name="doc_type",
                    message="Select doc_type(s) or press Enter to skip:",
                    choices=["Journal", "Book"],
                ),
            ]
            answers = inquirer.prompt(doc_type_choices)
            doc_type = ",".join(answers["doc_type"])

            search_parameters = {
                "subject": subject,
                "keyword": keyword,
                "language": language,
                "year": year,
                "type": doc_type,
            }

        return search_parameters

    def build_query(self, search_parameters: dict) -> str:
        """build API query"""

        if "query" in search_parameters:
            query = search_parameters["query"]

        else:
            constraints = []
            for key, value in search_parameters.items():
                if value == "":
                    continue

                if key in ["subject", "language", "doc_type"]:

                    subject_query = (
                        f'{key}:"'
                        + f'" OR {key}:"'.join(value for value in value.split(","))
                        + '"'
                    )
                    constraints.append(f"({subject_query})")
                    continue

                if value:
                    constraints.append(f'{key}:"{value}"')

            query = " AND ".join(constraints)

        return urllib.parse.quote(query)

    def _build_api_search_url(self, query: str, api_key: str, start: int = 1) -> str:
        return f"https://api.springernature.com/meta/v2/json?q={query}&api_key={api_key}&s={start}"

    def get_query_return(self) -> typing.Iterator[colrev.record.record.Record]:
        """Get the records from a API search"""
        query = self.build_query(self.search_source.search_parameters)
        api_key = self.get_api_key()
        start = 1

        while True:
            full_url = self._build_api_search_url(
                query=query, api_key=api_key, start=start
            )
            response = requests.get(full_url, timeout=10)
            if response.status_code != 200:
                print(
                    f"Error - API search failed for the following reason: {response.status_code}"
                )
                return

            data = response.json()

            for record in data.get("records", []):
                yield self._create_record(record)

            next_page = data.get("nextPage")
            if not next_page:
                break

            start_str = next_page.split("s=")[1].split("&")[0]
            start = int(start_str)

    def _run_api_search(
        self, springer_feed: colrev.ops.search_api_feed.SearchAPIFeed, rerun: bool
    ) -> None:
        """save API search results"""
        for record in self.get_query_return():
            springer_feed.add_update_record(record)

        springer_feed.save()

    def _create_record(self, doc: dict) -> colrev.record.record.Record:
        """Fieldmapper API search"""
        record_dict = {Fields.ID: doc["identifier"]}
        record_dict[Fields.ENTRYTYPE] = "misc"

        if "Article" in doc["contentType"]:
            record_dict[Fields.ENTRYTYPE] = "article"
        elif "Chapter" in doc["contentType"]:
            record_dict[Fields.ENTRYTYPE] = "inproceedings"
        elif "Book" in doc["contentType"]:
            record_dict[Fields.ENTRYTYPE] = "book"

        record_dict.update(
            {
                Fields.AUTHOR: " and ".join(
                    creator.get("creator", "") for creator in doc.get("creators", [])
                ),
                Fields.TITLE: doc.get("title", ""),
                Fields.PUBLISHER: doc.get("publisher", ""),
                Fields.BOOKTITLE: (
                    doc.get("publicationName", "")
                    if doc.get("publicationType") == "Book"
                    else ""
                ),
                Fields.JOURNAL: (
                    doc.get("publicationName", "")
                    if doc.get("publicationType") == "Journal"
                    else ""
                ),
                Fields.YEAR: (
                    doc.get("publicationDate", "").split("-")[0]
                    if doc.get("publicationDate")
                    else ""
                ),
                Fields.VOLUME: doc.get("volume", ""),
                Fields.NUMBER: doc.get("number", ""),
                Fields.PAGES: (
                    f"{doc.get('startingPage', '')}-{doc.get('endingPage', '')}"
                    if doc.get("startingPage") and doc.get("endingPage")
                    else ""
                ),
                Fields.DOI: doc.get("doi", ""),
                Fields.URL: next(
                    (
                        url.get("value", "")
                        for url in doc.get("url", [])
                        if url.get("format") == "html"
                    ),
                    doc.get("url", [{}])[0].get("value", "") if doc.get("url") else "",
                ),
            }
        )

        record = colrev.record.record.Record(data=record_dict)
        if Fields.LANGUAGE in record.data:
            try:
                record.data[Fields.LANGUAGE] = record.data[Fields.LANGUAGE][0]
                self.language_service.unify_to_iso_639_3_language_codes(record=record)
            except colrev_exceptions.InvalidLanguageCodeException:
                del record.data[Fields.LANGUAGE]
        return record

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def _load_csv(self) -> dict:
        def entrytype_setter(record_dict: dict) -> None:
            if record_dict["Content Type"] == "Article":
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE
            elif record_dict["Content Type"] == "Book":
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.BOOK
            elif record_dict["Content Type"] == "Chapter":
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.INBOOK
            else:
                record_dict[Fields.ENTRYTYPE] = "misc"

        def field_mapper(record_dict: dict) -> None:
            record_dict[Fields.TITLE] = record_dict.pop("Item Title", "")
            if record_dict[Fields.ENTRYTYPE] == ENTRYTYPES.ARTICLE:
                record_dict[Fields.JOURNAL] = record_dict.pop("Publication Title", "")
            elif record_dict[Fields.ENTRYTYPE] == ENTRYTYPES.BOOK:
                record_dict[Fields.BOOKTITLE] = record_dict.pop("Book Series Title", "")
            elif record_dict[Fields.ENTRYTYPE] == ENTRYTYPES.INBOOK:
                record_dict[Fields.CHAPTER] = record_dict.pop("Item Title", "")
                record_dict[Fields.TITLE] = record_dict.pop("Publication Title", "")
            record_dict[Fields.VOLUME] = record_dict.pop("Journal Volume", "")
            record_dict[Fields.NUMBER] = record_dict.pop("Journal Issue", "")
            record_dict[Fields.DOI] = record_dict.pop("Item DOI", "")
            record_dict[Fields.AUTHOR] = record_dict.pop("Authors", "")
            record_dict[Fields.YEAR] = record_dict.pop("Publication Year", "")
            record_dict[Fields.URL] = record_dict.pop("URL", "")
            record_dict.pop("Content Type", None)

            # Fix authors
            # pylint: disable=colrev-missed-constant-usage
            if Fields.AUTHOR in record_dict:
                # a-bd-z: do not match McDonald
                record_dict[Fields.AUTHOR] = re.sub(
                    r"([a-bd-z]{1})([A-Z]{1})",
                    r"\g<1> and \g<2>",
                    record_dict["author"],
                )

            for key in list(record_dict.keys()):
                value = record_dict[key]
                record_dict[key] = str(value)
                if value == "" or pd.isna(value):
                    del record_dict[key]

        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            unique_id_field="Item DOI",
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=self.review_manager.logger,
        )
        return records

    def api_ui(self) -> None:
        """User API key insertion"""

        api_key = self.get_api_key()

        if api_key:
            print("Your API key is available\n")

            change_api_key = [
                inquirer.List(
                    name="change_api_key",
                    message="Do you want to change your saved API key?",
                    choices=["no", "yes"],
                ),
            ]

            answers = inquirer.prompt(change_api_key)

            if answers["change_api_key"] == "no":
                return

        print("\n An API key is required for search \n\n")
        self._api_key_ui()

    def get_api_key(self) -> str:
        """Get API key from settings"""
        api_key = self.review_manager.environment_manager.get_settings_by_key(
            self.SETTINGS["api_key"]
        )
        if api_key:
            return api_key
        return ""

    def _is_springer_link_api_key(self, previous: dict, answer: str) -> bool:
        """Validate SpringerLink API key format"""
        api_key_pattern = re.compile(r"[a-z0-9]{32}")
        if not api_key_pattern.fullmatch(answer):
            raise inquirer.errors.ValidationError(
                "", reason="Invalid SpringerLink API key format."
            )

        full_url = self._build_api_search_url(
            query="doi:10.1007/978-3-319-07410-8_4", api_key=answer
        )
        response = requests.get(full_url, timeout=10)
        if response.status_code != 200:
            raise inquirer.errors.ValidationError("", reason="Error: Invalid API key.")
        print(
            f"\n{Colors.GREEN}Successfully authenticated with Springer Link API{Colors.END}"
        )
        return True

    def _is_year(self, previous: dict, answer: str) -> bool:
        """Validate year format"""
        year_pattern = re.compile(r"\d{4}")
        if not year_pattern.fullmatch(answer) and not answer == "":
            raise inquirer.errors.ValidationError("", reason="Invalid year format.")
        return True

    def _api_key_ui(self) -> None:
        """User Interface to enter API key"""

        questions = [
            inquirer.Text(
                "github_api_key",
                message="Enter your Springer Link API key",
                validate=self._is_springer_link_api_key,
            ),
        ]
        answers = inquirer.prompt(questions)
        input_key = answers["github_api_key"]
        self.review_manager.environment_manager.update_registry(
            self.SETTINGS["api_key"], input_key
        )

    def _load_bib(self) -> dict:
        """load bib file"""
        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            logger=self.review_manager.logger,
            unique_id_field="ID",
        )
        return records

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".csv":
            return self._load_csv()

        if self.search_source.filename.suffix == ".bib":
            return self._load_bib()

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for Springer Link"""

        return record



#! /usr/bin/env python
"""SearchSource: GoogleScholar"""
from __future__ import annotations

from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class GoogleScholarSearchSource:
    """GoogleScholar"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    endpoint = "colrev.google_scholar"
    # pylint: disable=colrev-missed-constant-usage
    source_identifier = "url"
    search_types = [SearchType.DB]

    ci_supported: bool = Field(default=False)
    heuristic_status = SearchSourceHeuristicStatus.supported

    db_url = "https://scholar.google.de/"

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.search_source = self.settings_class(**settings)
        self.source_operation = source_operation
        self.review_manager = source_operation.review_manager

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for GoogleScholar"""

        result = {"confidence": 0.0}
        if data.count("https://scholar.google.com/scholar?q=relat") > 0.9 * data.count(
            "\n@"
        ):
            result["confidence"] = 1.0
            return result

        if data.count("{pop0") > 0.9 * data.count("\n@"):
            result["confidence"] = 1.0
            return result

        if (
            data.count("https://scholar.google.com/scholar?q=relat")
            == data.count('"uid": "GS:')
            and data.count('"uid": "GS:') > 0
        ):
            result["confidence"] = 1.0
            return result

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        search_source = operation.create_db_source(
            search_source_cls=cls,
            params={},
        )
        operation.add_source_and_search(search_source)
        return search_source

    def search(self, rerun: bool) -> None:
        """Run a search of GoogleScholar"""

        if self.search_source.search_type == SearchType.DB:
            self.source_operation.run_db_search(  # type: ignore
                search_source_cls=self.__class__,
                source=self.search_source,
            )
            return

        raise NotImplementedError

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":

            def bib_field_mapper(record_dict: dict) -> None:
                if "related" in record_dict:
                    record_dict[f"{self.endpoint}.related"] = record_dict.pop("related")
                if "note" in record_dict:
                    record_dict[f"{self.endpoint}.note"] = record_dict.pop("note")
                if "type" in record_dict:
                    record_dict[f"{self.endpoint}.type"] = record_dict.pop("type")

            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                field_mapper=bib_field_mapper,
                logger=self.review_manager.logger,
            )
            return records

        if self.search_source.filename.suffix == ".json":
            # pylint: disable=too-many-branches
            def json_field_mapper(record_dict: dict) -> None:
                if "related" in record_dict:
                    record_dict[f"{self.endpoint}.related"] = record_dict.pop("related")
                if "note" in record_dict:
                    record_dict[f"{self.endpoint}.note"] = record_dict.pop("note")
                if "type" in record_dict:
                    record_dict[f"{self.endpoint}.type"] = record_dict.pop("type")
                if "article_url" in record_dict:
                    record_dict[f"{self.endpoint}.article_url"] = record_dict.pop(
                        "article_url"
                    )
                if "cites_url" in record_dict:
                    record_dict[f"{self.endpoint}.cites_url"] = record_dict.pop(
                        "cites_url"
                    )
                if "related_url" in record_dict:
                    record_dict[f"{self.endpoint}.related_url"] = record_dict.pop(
                        "related_url"
                    )
                if "fulltext_url" in record_dict:
                    record_dict[f"{self.endpoint}.fulltext_url"] = record_dict.pop(
                        "fulltext_url"
                    )

                if "uid" in record_dict:
                    record_dict[f"{self.endpoint}.uid"] = record_dict.pop("uid")
                if "source" in record_dict:
                    record_dict[Fields.JOURNAL] = record_dict.pop("source")
                if "cites" in record_dict:
                    record_dict[f"{self.endpoint}.cites"] = record_dict.pop("cites")

                record_dict.pop("volume", None)
                record_dict.pop("issue", None)
                record_dict.pop("startpage", None)
                record_dict.pop("endpage", None)
                record_dict.pop("ecc", None)
                record_dict.pop("use", None)
                record_dict.pop("rank", None)
                if "authors" in record_dict:
                    authors = record_dict.pop("authors")
                    for i, author in enumerate(authors):
                        names = author.split(" ")
                        if len(names) > 1:
                            first_name, last_name = names[0], " ".join(names[1:])
                            authors[i] = f"{last_name}, {first_name}"
                        else:
                            authors[i] = names[0]

                    record_dict[Fields.AUTHOR] = " and ".join(authors)

                for key, value in record_dict.items():
                    record_dict[key] = str(value)

            def json_entrytype_setter(record_dict: dict) -> None:
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.MISC

            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                entrytype_setter=json_entrytype_setter,
                field_mapper=json_field_mapper,
                # Note: uid not always available.
                unique_id_field="INCREMENTAL",
                logger=self.review_manager.logger,
            )

            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for GoogleScholar"""
        if "cites: https://scholar.google.com/scholar?cites=" in record.data.get(
            "note", ""
        ):
            note = record.data["note"]
            source_field = record.get_field_provenance_source("note")
            record.rename_field(key="note", new_key=Fields.CITED_BY)
            record.update_field(
                key=Fields.CITED_BY,
                value=record.data[Fields.CITED_BY][
                    : record.data[Fields.CITED_BY].find(" cites: ")
                ],
                source="replace_link",
            )
            record.update_field(
                key="cited_by_link",
                value=note[note.find("cites: ") + 7 :],
                append_edit=False,
                source=source_field + "|extract-from-note",
            )
        if Fields.ABSTRACT in record.data:
            # Note: abstracts provided by GoogleScholar are very incomplete
            record.remove_field(key=Fields.ABSTRACT)

        return record



#!/usr/bin/env python3
"""The CoLRev review manager (main entrypoint)."""
from __future__ import annotations

import logging
import os
import pprint
import typing
from datetime import timedelta
from pathlib import Path

import git
import requests_cache
import yaml

import colrev.dataset
import colrev.exceptions as colrev_exceptions
import colrev.logger
import colrev.ops.check
import colrev.ops.checker
import colrev.process.operation
import colrev.record.qm.quality_model
import colrev.settings
from colrev.constants import Colors
from colrev.constants import Filepaths
from colrev.constants import OperationsType
from colrev.paths import PathManager


class ReviewManager:
    """Class for managing individual CoLRev review project (repositories)"""

    # pylint: disable=import-outside-toplevel
    # pylint: disable=redefined-outer-name
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-public-methods
    # pylint: disable=too-many-arguments

    notified_next_operation = None
    """ReviewManager was notified for the upcoming process and
    will provide access to the Dataset"""

    dataset: colrev.dataset.Dataset
    """The review dataset object"""

    path: Path
    """Path of the project repository"""

    shell_mode = False

    def __init__(
        self,
        *,
        path_str: typing.Optional[str] = None,
        force_mode: bool = False,
        verbose_mode: bool = False,
        high_level_operation: bool = False,
        navigate_to_home_dir: bool = True,
        exact_call: str = "",
        skip_upgrade: bool = True,
    ) -> None:
        self.force_mode = force_mode
        """Force mode variable (bool)"""
        self.verbose_mode = verbose_mode
        """Verbose mode variable (bool)"""
        self.high_level_operation = high_level_operation
        """A high-level operation was called (bool)"""
        # Note : mostly for formatting output

        if navigate_to_home_dir:
            self.path = self._get_project_home_dir(path_str=path_str)
        else:
            self.path = Path.cwd()

        self.paths = PathManager(self.path)

        self.exact_call = exact_call

        try:
            if self.paths.settings.is_file():
                self.paths.data.mkdir(parents=True, exist_ok=True)
                self.paths.search.mkdir(parents=True, exist_ok=True)
                self.paths.pdf.mkdir(parents=True, exist_ok=True)
                self.paths.output.mkdir(parents=True, exist_ok=True)

            report_logger, logger = self.get_loggers()
            self.report_logger = report_logger
            self.logger = logger

            self.environment_manager = self.get_environment_manager()

            self.p_printer = pprint.PrettyPrinter(indent=4, width=140, compact=False)
            # run update before settings/data (which may require changes/fail without update)
            if not skip_upgrade:  # pragma: no cover
                self._check_update()
            self.settings = self.load_settings()
            self.dataset = colrev.dataset.Dataset(review_manager=self)

        except Exception as exc:  # pylint: disable=broad-except

            if (self.path / Path(".git")).is_dir():
                if git.Repo().active_branch.name == "gh-pages":  # pragma: no cover
                    raise colrev_exceptions.RepoSetupError(
                        msg="Currently on gh-pages branch. Switch to main: "
                        + f"{Colors.ORANGE}git switch main{Colors.END}"
                    )

            if not force_mode:
                raise exc

    # pylint: disable=too-many-arguments
    def update_config(
        self,
        *,
        force_mode: bool = False,
        verbose_mode: bool = False,
        high_level_operation: bool = False,
        exact_call: str = "",
    ) -> None:
        """Update review_manager's state"""
        self.force_mode = force_mode
        self.verbose_mode = verbose_mode
        self.high_level_operation = high_level_operation
        self.exact_call = exact_call
        report_logger, logger = self.get_loggers()
        self.report_logger = report_logger
        self.logger = logger

    def get_loggers(self) -> typing.Tuple[logging.Logger, logging.Logger]:
        """return loggers"""
        if self.verbose_mode:
            return colrev.logger.setup_report_logger(
                review_manager=self, level=logging.DEBUG
            ), colrev.logger.setup_logger(review_manager=self, level=logging.DEBUG)
        return colrev.logger.setup_report_logger(
            review_manager=self, level=logging.INFO
        ), colrev.logger.setup_logger(review_manager=self, level=logging.INFO)

    def _check_update(self) -> None:
        # Once the following has run for all repositories,
        # it should only be called when the versions differ.
        # last_version, current_version = self.get_colrev_versions()
        # if last_version != current_version:
        upgrade_operation = self.get_upgrade()
        upgrade_operation.main()

    def get_committer(self) -> typing.Tuple[str, str]:
        """Get the committer name and email"""
        return self.environment_manager.get_name_mail_from_git()

    def _get_project_home_dir(self, *, path_str: typing.Optional[str] = None) -> Path:
        if path_str:
            original_dir = Path(path_str)
        else:
            original_dir = Path.cwd()

        while ".git" not in [f.name for f in original_dir.iterdir() if f.is_dir()]:
            if original_dir.parent == original_dir:  # reached root
                break
            original_dir = original_dir.parent

        if original_dir.parent == original_dir:  # reached root
            raise colrev.exceptions.RepoSetupError(
                "Failed to locate a .git directory. "
                "Ensure you are within a Git repository, "
                "or set navigate_to_home_dir=False for init."
            )

        return original_dir

    def load_settings(self) -> colrev.settings.Settings:
        """Load the settings"""
        self.settings = colrev.settings.load_settings(settings_path=self.paths.settings)
        return self.settings

    def save_settings(self) -> None:
        """Save the settings"""
        colrev.settings.save_settings(review_manager=self)

    def reset_report_logger(self) -> None:
        """Reset the report logger"""
        colrev.logger.reset_report_logger(review_manager=self)

    def check_repo(self) -> dict:
        """Check the repository"""
        checker = colrev.ops.checker.Checker(review_manager=self)
        return checker.check_repo()

    def in_virtualenv(self) -> bool:  # pragma: no cover
        """Check whether CoLRev operates in a virtual environment"""
        return colrev.ops.checker.Checker.in_virtualenv()

    def check_repository_setup(self) -> None:
        """Check the repository setup"""
        checker = colrev.ops.checker.Checker(review_manager=self)
        checker.check_repository_setup()

    def get_colrev_versions(self) -> list[str]:
        """Get the CoLRev versions"""
        checker = colrev.ops.checker.Checker(review_manager=self)
        return checker.get_colrev_versions()

    def report(self, *, msg_file: Path) -> None:
        """Append commit-message report if not already available
        (Entrypoint for pre-commit hooks)
        """
        import colrev.ops.commit
        import colrev.ops.correct

        with open(msg_file, encoding="utf8") as file:
            available_contents = file.read()

        with open(msg_file, "w", encoding="utf8") as file:
            file.write(available_contents)
            # Don't append if it's already there
            # update = False
            # if "Command" not in available_contents:
            #     update = True
            # if "Properties" in available_contents:
            #     update = False
            # if update:
            commit = colrev.ops.commit.Commit(
                review_manager=self,
                msg=available_contents,
                manual_author=True,
                script_name="MANUAL",
            )
            commit.update_report(msg_file=msg_file)

        if (
            not self.settings.is_curated_masterdata_repo()
            and self.dataset.records_changed()
        ):  # pragma: no cover
            colrev.ops.check.CheckOperation(self)  # to notify
            corrections_operation = colrev.ops.correct.Corrections(review_manager=self)
            corrections_operation.check_corrections_of_records()

    def sharing(self) -> dict:
        """Check whether sharing requirements are met
        (Entrypoint for pre-commit hooks)
        """

        self.notified_next_operation = OperationsType.check
        advisor = self.get_advisor()
        sharing_advice = advisor.get_sharing_instructions()
        return sharing_advice

    def update_status_yaml(
        self, *, add_to_git: bool = True, records: typing.Optional[dict] = None
    ) -> None:
        """Update the STATUS_FILE"""

        status_stats = self.get_status_stats(records=records)
        exported_dict = status_stats.model_dump()
        exported_dict.pop("origin_states_dict")
        exported_dict.pop("perc_curated")
        exported_dict.pop("screening_statistics")
        exported_dict.pop("nr_origins")
        with open(self.paths.status, "w", encoding="utf8") as file:
            yaml.dump(exported_dict, file, allow_unicode=True)
        if add_to_git:
            self.dataset.add_changes(self.paths.STATUS_FILE)

    def get_upgrade(self) -> colrev.ops.upgrade.Upgrade:  # pragma: no cover
        """Get an upgrade object"""

        import colrev.ops.upgrade

        return colrev.ops.upgrade.Upgrade(review_manager=self)

    def get_repare(self) -> colrev.ops.repare.Repare:  # pragma: no cover
        """Get a a repare object"""

        import colrev.ops.repare

        return colrev.ops.repare.Repare(review_manager=self)

    def get_remove_operation(self) -> colrev.ops.remove.Remove:  # pragma: no cover
        """Get a a remove object"""

        import colrev.ops.remove

        return colrev.ops.remove.Remove(review_manager=self)

    def get_merge_operation(self) -> colrev.ops.merge.Merge:  # pragma: no cover
        """Get a merge object"""

        import colrev.ops.merge

        return colrev.ops.merge.Merge(review_manager=self)

    def get_advisor(self) -> colrev.ops.advisor.Advisor:  # pragma: no cover
        """Get an advisor object"""

        import colrev.ops.advisor

        return colrev.ops.advisor.Advisor(review_manager=self)

    def get_checker(self) -> colrev.ops.checker.Checker:  # pragma: no cover
        """Get a checker object"""

        return colrev.ops.checker.Checker(review_manager=self)

    def get_qm(self) -> colrev.record.qm.quality_model.QualityModel:  # pragma: no cover
        """Get the quality model"""

        return colrev.record.qm.quality_model.QualityModel(
            defects_to_ignore=self.settings.prep.defects_to_ignore
        )

    def get_pdf_qm(
        self,
    ) -> colrev.record.qm.quality_model.QualityModel:  # pragma: no cover
        """Get the PDF quality model"""

        return colrev.record.qm.quality_model.QualityModel(
            defects_to_ignore=self.settings.pdf_get.defects_to_ignore,
            pdf_mode=True,
            path=self.path,
        )

    def get_status_stats(
        self, *, records: typing.Optional[dict] = None
    ) -> colrev.process.status.StatusStats:  # pragma: no cover
        """Get a status stats object"""

        import colrev.process.status

        colrev.ops.check.CheckOperation(self)

        if records is None:
            records = self.dataset.load_records_dict()
        return colrev.process.status.get_status_stats(
            review_manager=self, records=records
        )

    def get_completeness_condition(self) -> bool:
        """Get the completeness condition"""
        status_stats = self.get_status_stats()
        return status_stats.completeness_condition

    @classmethod
    def get_package_manager(
        cls,
    ) -> colrev.package_manager.package_manager.PackageManager:  # pragma: no cover
        """Get a package manager object"""

        import colrev.package_manager.package_manager

        return colrev.package_manager.package_manager.PackageManager()

    @classmethod
    def get_grobid_service(
        cls,
    ) -> colrev.env.grobid_service.GrobidService:  # pragma: no cover
        """Get a grobid service object"""
        import colrev.env.grobid_service

        environment_manager = cls.get_environment_manager()
        return colrev.env.grobid_service.GrobidService(
            environment_manager=environment_manager
        )

    def get_tei(
        self,
        *,
        pdf_path: typing.Optional[Path] = None,
        tei_path: typing.Optional[Path] = None,
    ) -> colrev.env.tei_parser.TEIParser:  # type: ignore # pragma: no cover
        """Get a tei object"""

        import colrev.env.tei_parser

        return colrev.env.tei_parser.TEIParser(
            environment_manager=self.environment_manager,
            pdf_path=self.path / pdf_path if pdf_path else None,
            tei_path=self.path / tei_path if tei_path else None,
        )

    @classmethod
    def get_environment_manager(
        cls,
    ) -> colrev.env.environment_manager.EnvironmentManager:  # pragma: no cover
        """Get an environment manager"""
        import colrev.env.environment_manager

        return colrev.env.environment_manager.EnvironmentManager()

    @classmethod
    def get_cached_session(cls) -> requests_cache.CachedSession:  # pragma: no cover
        """Get a cached session"""

        return requests_cache.CachedSession(
            str(Filepaths.PREP_REQUESTS_CACHE_FILE),
            backend="sqlite",
            expire_after=timedelta(days=30),
        )

    @classmethod
    def get_resources(cls) -> colrev.env.resources.Resources:  # pragma: no cover
        """Get a resources object"""
        import colrev.env.resources

        return colrev.env.resources.Resources()

    def get_search_operation(
        self, *, notify_state_transition_operation: bool = True
    ) -> colrev.ops.search.Search:  # pragma: no cover
        """Get a search operation object"""
        import colrev.ops.search

        return colrev.ops.search.Search(
            review_manager=self,
            notify_state_transition_operation=notify_state_transition_operation,
        )

    def get_load_operation(
        self,
        notify_state_transition_operation: bool = True,
        hide_load_explanation: bool = False,
    ) -> colrev.ops.load.Load:  # pragma: no cover
        """Get a load operation object"""
        import colrev.ops.load

        return colrev.ops.load.Load(
            review_manager=self,
            notify_state_transition_operation=notify_state_transition_operation,
            hide_load_explanation=hide_load_explanation,
        )

    def get_prep_operation(
        self,
        *,
        notify_state_transition_operation: bool = True,
        polish: bool = False,
        cpu: int = 4,
        debug: bool = False,
    ) -> colrev.ops.prep.Prep:  # pragma: no cover
        """Get a prep operation object"""
        if debug:
            import colrev.ops.prep_debug

            return colrev.ops.prep_debug.PrepDebug(
                review_manager=self,
                notify_state_transition_operation=notify_state_transition_operation,
                polish=polish,
            )

        import colrev.ops.prep

        return colrev.ops.prep.Prep(
            review_manager=self,
            notify_state_transition_operation=notify_state_transition_operation,
            polish=polish,
            cpu=cpu,
        )

    def get_prep_man_operation(
        self, *, notify_state_transition_operation: bool = True
    ) -> colrev.ops.prep_man.PrepMan:  # pragma: no cover
        """Get a prep-man operation object"""
        import colrev.ops.prep_man

        return colrev.ops.prep_man.PrepMan(
            review_manager=self,
            notify_state_transition_operation=notify_state_transition_operation,
        )

    def get_dedupe_operation(
        self, *, notify_state_transition_operation: bool = True
    ) -> colrev.ops.dedupe.Dedupe:  # pragma: no cover
        """Get a dedupe operation object"""
        import colrev.ops.dedupe

        return colrev.ops.dedupe.Dedupe(
            review_manager=self,
            notify_state_transition_operation=notify_state_transition_operation,
        )

    def get_prescreen_operation(
        self, *, notify_state_transition_operation: bool = True
    ) -> colrev.ops.prescreen.Prescreen:  # pragma: no cover
        """Get a prescreen operation object"""

        import colrev.ops.prescreen

        return colrev.ops.prescreen.Prescreen(
            review_manager=self,
            notify_state_transition_operation=notify_state_transition_operation,
        )

    def get_pdf_get_operation(
        self, *, notify_state_transition_operation: bool = True
    ) -> colrev.ops.pdf_get.PDFGet:  # pragma: no cover
        """Get a pdf-get operation object"""
        import colrev.ops.pdf_get

        return colrev.ops.pdf_get.PDFGet(
            review_manager=self,
            notify_state_transition_operation=notify_state_transition_operation,
        )

    def get_pdf_get_man_operation(
        self, *, notify_state_transition_operation: bool = True
    ) -> colrev.ops.pdf_get_man.PDFGetMan:  # pragma: no cover
        """Get a pdf-get-man operation object"""
        import colrev.ops.pdf_get_man

        return colrev.ops.pdf_get_man.PDFGetMan(
            review_manager=self,
            notify_state_transition_operation=notify_state_transition_operation,
        )

    def get_pdf_prep_operation(
        self, *, reprocess: bool = False, notify_state_transition_operation: bool = True
    ) -> colrev.ops.pdf_prep.PDFPrep:  # pragma: no cover
        """Get a pdfprep operation object"""
        import colrev.ops.pdf_prep

        return colrev.ops.pdf_prep.PDFPrep(
            review_manager=self,
            reprocess=reprocess,
            notify_state_transition_operation=notify_state_transition_operation,
        )

    def get_pdf_prep_man_operation(
        self, *, notify_state_transition_operation: bool = True
    ) -> colrev.ops.pdf_prep_man.PDFPrepMan:  # pragma: no cover
        """Get a pdf-prep-man operation object"""
        import colrev.ops.pdf_prep_man

        return colrev.ops.pdf_prep_man.PDFPrepMan(
            review_manager=self,
            notify_state_transition_operation=notify_state_transition_operation,
        )

    def get_screen_operation(
        self, *, notify_state_transition_operation: bool = True
    ) -> colrev.ops.screen.Screen:  # pragma: no cover
        """Get a screen operation object"""
        import colrev.ops.screen

        return colrev.ops.screen.Screen(
            review_manager=self,
            notify_state_transition_operation=notify_state_transition_operation,
        )

    def get_data_operation(
        self, *, notify_state_transition_operation: bool = True
    ) -> colrev.ops.data.Data:  # pragma: no cover
        """Get a data operation object"""
        import colrev.ops.data

        return colrev.ops.data.Data(
            review_manager=self,
            notify_state_transition_operation=notify_state_transition_operation,
        )

    def get_status_operation(self) -> colrev.ops.status.Status:  # pragma: no cover
        """Get a status operation object"""

        import colrev.ops.status

        return colrev.ops.status.Status(review_manager=self)

    def get_validate_operation(
        self,
    ) -> colrev.ops.validate.Validate:  # pragma: no cover
        """Get a validate operation object"""
        import colrev.ops.validate

        return colrev.ops.validate.Validate(review_manager=self)

    def get_trace_operation(self) -> colrev.ops.trace.Trace:  # pragma: no cover
        """Get a trace operation object"""
        import colrev.ops.trace

        return colrev.ops.trace.Trace(review_manager=self)

    def get_distribute_operation(
        self,
    ) -> colrev.ops.distribute.Distribute:  # pragma: no cover
        """Get a distribute operation object"""

        import colrev.ops.distribute

        return colrev.ops.distribute.Distribute(review_manager=self)

    # pylint: disable=line-too-long
    def get_push_operation(self, **kwargs) -> colrev.ops.push.Push:  # type: ignore # pragma: no cover
        """Get a push operation object"""

        import colrev.ops.push

        return colrev.ops.push.Push(review_manager=self, **kwargs)

    def get_pull_operation(self) -> colrev.ops.pull.Pull:  # pragma: no cover
        """Get a pull operation object"""

        import colrev.ops.pull

        return colrev.ops.pull.Pull(review_manager=self)

    def get_connecting_review_manager(
        self,
        *,
        path_str: typing.Optional[str] = None,
        force_mode: bool = False,
        verbose_mode: bool = False,
    ) -> ReviewManager:  # pragma: no cover
        """Get a (connecting) ReviewManager object for another CoLRev repository"""
        return type(self)(
            path_str=path_str, force_mode=force_mode, verbose_mode=verbose_mode
        )

    @classmethod
    def in_test_environment(cls) -> bool:
        """Check whether CoLRev runs in a test environment"""

        return "pytest" in os.getcwd()

    @classmethod
    def in_ci_environment(
        cls,
    ) -> bool:
        """Check whether CoLRev runs in a continuous-integration environment"""

        identifier_list = [
            "GITHUB_ACTIONS",
            "CIRCLECI",
            "TRAVIS",
            "GITLAB_CI",
        ]
        return any("true" == os.getenv(x) for x in identifier_list)



#! /usr/bin/env python
"""Retrieval of PDFs from the LocalIndex"""
from __future__ import annotations

import shutil
from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.env.local_index
import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields

# pylint: disable=duplicate-code
# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PDFGetInterface)
class LocalIndexPDFGet:
    """Get PDFs from LocalIndex"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=False)

    def __init__(
        self,
        *,
        pdf_get_operation: colrev.ops.pdf_get.PDFGet,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.pdf_get_operation = pdf_get_operation
        self.review_manager = pdf_get_operation.review_manager

    def get_pdf(
        self, record: colrev.record.record.Record
    ) -> colrev.record.record.Record:
        """Get PDFs from the local-index"""

        local_index = colrev.env.local_index.LocalIndex(
            verbose_mode=self.review_manager.verbose_mode
        )

        try:
            retrieved_record = local_index.retrieve(record.data, include_file=True)
        except colrev_exceptions.RecordNotInIndexException:
            return record

        if Fields.FILE in retrieved_record.data:
            record.update_field(
                key=Fields.FILE,
                value=str(retrieved_record.data[Fields.FILE]),
                source="local_index",
            )
            self.pdf_get_operation.import_pdf(record)
            if Fields.FULLTEXT in retrieved_record.data:
                try:
                    record.get_tei_filename().write_text(
                        retrieved_record.data[Fields.FULLTEXT]
                    )
                except FileNotFoundError:
                    pass
                del retrieved_record.data[Fields.FULLTEXT]
            else:
                tei_ext_path = Path(
                    retrieved_record.data[Fields.FILE]
                    .replace("pdfs/", ".tei/")
                    .replace(".pdf", ".tei.xml")
                )
                if tei_ext_path.is_file():
                    new_path = record.get_tei_filename()
                    new_path.resolve().parent.mkdir(exist_ok=True, parents=True)
                    shutil.copy(tei_ext_path, new_path)

        return record



#! /usr/bin/env python
"""SearchSource: Unknown source (default for all other sources)"""
from __future__ import annotations

import re
import shutil
from pathlib import Path

import pandas as pd
import zope.interface
from pydantic import Field
from rapidfuzz import fuzz

import colrev.env.language_service
import colrev.exceptions as colrev_exceptions
import colrev.loader.load_utils
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
import colrev.record.record_prep
from colrev.constants import Colors
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import FieldSet
from colrev.constants import FieldValues
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class UnknownSearchSource:
    """Unknown SearchSource"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    endpoint = "colrev.unknown_source"

    source_identifier = "colrev.unknown_source"
    search_types = [
        SearchType.DB,
        SearchType.OTHER,
        SearchType.BACKWARD_SEARCH,
        SearchType.FORWARD_SEARCH,
        SearchType.TOC,
    ]

    ci_supported: bool = Field(default=False)
    heuristic_status = SearchSourceHeuristicStatus.na

    db_url = ""

    HTML_CLEANER = re.compile("<.*?>")
    _padding = 40

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.search_source = self.settings_class(**settings)
        self.review_manager = source_operation.review_manager
        self.language_service = colrev.env.language_service.LanguageService()
        self.operation = source_operation

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for unknown sources"""

        result = {"confidence": 0.1}

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        params_dict = {}
        if params:
            for item in params.split(";"):
                key, value = item.split("=")
                params_dict[key] = value
        search_source = operation.create_db_source(
            search_source_cls=cls,
            params=params_dict,
        )
        operation.add_source_and_search(search_source)
        return search_source

    def search(self, rerun: bool) -> None:
        """Run a search of Crossref"""

        if self.search_source.search_type == SearchType.DB:
            self.operation.run_db_search(  # type: ignore
                search_source_cls=self.__class__, source=self.search_source
            )

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def _rename_erroneous_extensions(self) -> None:
        if self.search_source.filename.suffix in [".xls", ".xlsx"]:
            return
        data = self.search_source.filename.read_text(encoding="utf-8")
        # # Correct the file extension if necessary
        if re.findall(
            r"^%0", data, re.MULTILINE
        ) and self.search_source.filename.suffix not in [".enl"]:
            new_filename = self.search_source.filename.with_suffix(".enl")
            self.review_manager.logger.info(
                f"{Colors.GREEN}Rename to {new_filename} "
                f"(because the format is .enl){Colors.END}"
            )
            shutil.move(str(self.search_source.filename), str(new_filename))
            self.review_manager.dataset.add_changes(
                self.search_source.filename, remove=True
            )
            self.search_source.filename = new_filename
            self.review_manager.dataset.add_changes(new_filename)
            self.review_manager.dataset.create_commit(
                msg=f"Rename {self.search_source.filename}"
            )
            return

        if re.findall(
            r"^TI ", data, re.MULTILINE
        ) and self.search_source.filename.suffix not in [".ris"]:
            new_filename = self.search_source.filename.with_suffix(".ris")
            self.review_manager.logger.info(
                f"{Colors.GREEN}Rename to {new_filename} "
                f"(because the format is .ris){Colors.END}"
            )
            shutil.move(str(self.search_source.filename), str(new_filename))
            self.review_manager.dataset.add_changes(
                self.search_source.filename, remove=True
            )
            self.search_source.filename = new_filename
            self.review_manager.dataset.add_changes(new_filename)
            self.review_manager.dataset.create_commit(
                msg=f"Rename {self.search_source.filename}"
            )

    def _load_ris(self, *, load_operation: colrev.ops.load.Load) -> dict:
        def entrytype_setter(record_dict: dict) -> None:
            # Based on https://github.com/aurimasv/translators/wiki/RIS-Tag-Map
            reference_types = {
                "JOUR": ENTRYTYPES.ARTICLE,
                "JFULL": ENTRYTYPES.ARTICLE,
                "ABST": ENTRYTYPES.ARTICLE,
                "INPR": ENTRYTYPES.ARTICLE,  # inpress
                "CONF": ENTRYTYPES.INPROCEEDINGS,
                "CPAPER": ENTRYTYPES.INPROCEEDINGS,
                "THES": ENTRYTYPES.PHDTHESIS,
                "REPT": ENTRYTYPES.TECHREPORT,
                "RPRT": ENTRYTYPES.TECHREPORT,
                "CHAP": ENTRYTYPES.INBOOK,
                "BOOK": ENTRYTYPES.BOOK,
                "NEWS": ENTRYTYPES.MISC,
                "BLOG": ENTRYTYPES.MISC,
            }
            if record_dict["TY"] in reference_types:
                record_dict[Fields.ENTRYTYPE] = reference_types[record_dict["TY"]]
            else:
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.MISC

        def field_mapper(record_dict: dict) -> None:
            # Based on https://github.com/aurimasv/translators/wiki/RIS-Tag-Map
            key_maps = {
                ENTRYTYPES.ARTICLE: {
                    "PY": Fields.YEAR,
                    "AU": Fields.AUTHOR,
                    "TI": Fields.TITLE,
                    "T2": Fields.JOURNAL,
                    "AB": Fields.ABSTRACT,
                    "VL": Fields.VOLUME,
                    "IS": Fields.NUMBER,
                    "DO": Fields.DOI,
                    "PB": Fields.PUBLISHER,
                    "UR": Fields.URL,
                    "fulltext": Fields.FULLTEXT,
                    "PMID": Fields.PUBMED_ID,
                    "KW": Fields.KEYWORDS,
                    "SP": Fields.PAGES,
                },
                ENTRYTYPES.INPROCEEDINGS: {
                    "PY": Fields.YEAR,
                    "AU": Fields.AUTHOR,
                    "TI": Fields.TITLE,
                    # "secondary_title": Fields.BOOKTITLE,
                    "DO": Fields.DOI,
                    "UR": Fields.URL,
                    # "fulltext": Fields.FULLTEXT,
                    "PMID": Fields.PUBMED_ID,
                    "KW": Fields.KEYWORDS,
                    "SP": Fields.PAGES,
                },
                ENTRYTYPES.INBOOK: {
                    "PY": Fields.YEAR,
                    "AU": Fields.AUTHOR,
                    # "primary_title": Fields.CHAPTER,
                    # "secondary_title": Fields.TITLE,
                    "DO": Fields.DOI,
                    "PB": Fields.PUBLISHER,
                    # "edition": Fields.EDITION,
                    "UR": Fields.URL,
                    # "fulltext": Fields.FULLTEXT,
                    "KW": Fields.KEYWORDS,
                    "SP": Fields.PAGES,
                },
                ENTRYTYPES.BOOK: {
                    "PY": Fields.YEAR,
                    "AU": Fields.AUTHOR,
                    # "primary_title": Fields.CHAPTER,
                    # "secondary_title": Fields.TITLE,
                    "DO": Fields.DOI,
                    "PB": Fields.PUBLISHER,
                    # "edition": Fields.EDITION,
                    "UR": Fields.URL,
                    # "fulltext": Fields.FULLTEXT,
                    "KW": Fields.KEYWORDS,
                    "SP": Fields.PAGES,
                },
                ENTRYTYPES.PHDTHESIS: {
                    "PY": Fields.YEAR,
                    "AU": Fields.AUTHOR,
                    "TI": Fields.TITLE,
                    "UR": Fields.URL,
                },
                ENTRYTYPES.TECHREPORT: {
                    "PY": Fields.YEAR,
                    "AU": Fields.AUTHOR,
                    "TI": Fields.TITLE,
                    "UR": Fields.URL,
                    # "fulltext": Fields.FULLTEXT,
                    "KW": Fields.KEYWORDS,
                    "PB": Fields.PUBLISHER,
                    "SP": Fields.PAGES,
                },
                ENTRYTYPES.MISC: {
                    "PY": Fields.YEAR,
                    "AU": Fields.AUTHOR,
                    "TI": Fields.TITLE,
                    "UR": Fields.URL,
                    # "fulltext": Fields.FULLTEXT,
                    "KW": Fields.KEYWORDS,
                    "PB": Fields.PUBLISHER,
                    "SP": Fields.PAGES,
                },
            }

            key_map = key_maps[record_dict[Fields.ENTRYTYPE]]
            for ris_key in list(record_dict.keys()):
                if ris_key in key_map:
                    standard_key = key_map[ris_key]
                    record_dict[standard_key] = record_dict.pop(ris_key)

            if "SP" in record_dict and "EP" in record_dict:
                record_dict[Fields.PAGES] = (
                    f"{record_dict.pop('SP')}--{record_dict.pop('EP')}"
                )

            if Fields.AUTHOR in record_dict and isinstance(
                record_dict[Fields.AUTHOR], list
            ):
                record_dict[Fields.AUTHOR] = " and ".join(record_dict[Fields.AUTHOR])
            if Fields.EDITOR in record_dict and isinstance(
                record_dict[Fields.EDITOR], list
            ):
                record_dict[Fields.EDITOR] = " and ".join(record_dict[Fields.EDITOR])
            if Fields.KEYWORDS in record_dict and isinstance(
                record_dict[Fields.KEYWORDS], list
            ):
                record_dict[Fields.KEYWORDS] = ", ".join(record_dict[Fields.KEYWORDS])

            record_dict.pop("TY", None)
            record_dict.pop("Y2", None)
            record_dict.pop("DB", None)
            record_dict.pop("C1", None)
            record_dict.pop("T3", None)
            record_dict.pop("AD", None)
            record_dict.pop("CY", None)
            record_dict.pop("M3", None)
            record_dict.pop("EP", None)
            record_dict.pop("ER", None)

            for key, value in record_dict.items():
                record_dict[key] = str(value)

        load_operation.ensure_append_only(self.search_source.filename)
        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            unique_id_field="INCREMENTAL",
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=self.review_manager.logger,
        )
        return records

    def _load_bib(self, *, load_operation: colrev.ops.load.Load) -> dict:
        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            logger=self.review_manager.logger,
            unique_id_field="ID",
        )
        return records

    # pylint: disable=colrev-missed-constant-usage
    def _table_drop_fields(self, *, record_dict: dict) -> None:
        for key in list(record_dict.keys()):
            if record_dict[key] in [f"no {key}", "", "nan"]:
                del record_dict[key]
        if (
            record_dict.get("number_of_cited_references", "NA")
            == "no Number-of-Cited-References"
        ):
            del record_dict["number_of_cited_references"]
        if "no file" in record_dict.get("file_name", "NA"):
            del record_dict["file_name"]

        if record_dict.get("cited_by", "NA") in [
            "no Times-Cited",
        ]:
            del record_dict["cited_by"]

        if "author_count" in record_dict:
            del record_dict["author_count"]
        if "citation_key" in record_dict:
            del record_dict["citation_key"]

    def _load_table(self, *, load_operation: colrev.ops.load.Load) -> dict:
        def entrytype_setter(record_dict: dict) -> None:
            if "type" in record_dict:
                record_dict[Fields.ENTRYTYPE] = record_dict.pop("type")

            if Fields.ENTRYTYPE not in record_dict:
                if record_dict.get(Fields.JOURNAL, "") != "":
                    record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE
                elif record_dict.get(Fields.BOOKTITLE, "") != "":
                    record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.INPROCEEDINGS
                else:
                    record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.MISC

            if record_dict[Fields.ENTRYTYPE] == ENTRYTYPES.INPROCEEDINGS:
                if (
                    Fields.JOURNAL in record_dict
                    and Fields.BOOKTITLE not in record_dict
                ):
                    record_dict[Fields.BOOKTITLE] = record_dict.pop(Fields.JOURNAL)
            elif record_dict[Fields.ENTRYTYPE] == ENTRYTYPES.ARTICLE:
                if (
                    Fields.BOOKTITLE in record_dict
                    and Fields.JOURNAL not in record_dict
                ):
                    record_dict[Fields.JOURNAL] = record_dict.pop(Fields.BOOKTITLE)

        def field_mapper(record_dict: dict) -> None:
            # replace all keys with lower case
            record_dict = {k.lower(): v for k, v in record_dict.items()}

            if "id" in record_dict:
                record_dict[Fields.ID] = record_dict.pop("id")
            if "entrytype" in record_dict:
                record_dict[Fields.ENTRYTYPE] = record_dict.pop("entrytype")
            if "type" in record_dict:
                record_dict[Fields.ENTRYTYPE] = record_dict.pop("type")
            if "authors" in record_dict:
                record_dict[Fields.AUTHOR] = record_dict.pop("authors", "")
            if "publication" in record_dict:
                record_dict[Fields.JOURNAL] = record_dict.pop("publication", "")

            record_dict.pop("Category", None)
            record_dict.pop("Affiliations", None)

            if "issue" in record_dict and Fields.NUMBER not in record_dict:
                record_dict[Fields.NUMBER] = record_dict.pop("issue")
                if record_dict[Fields.NUMBER] == "no issue":
                    del record_dict[Fields.NUMBER]

            if "publication_year" in record_dict and Fields.YEAR not in record_dict:
                record_dict[Fields.YEAR] = record_dict.pop("publication_year")

            # Note: this is a simple heuristic:
            if (
                "journal/book" in record_dict
                and Fields.JOURNAL not in record_dict
                and Fields.DOI in record_dict
            ):
                record_dict[Fields.JOURNAL] = record_dict.pop("journal/book")

            if "author" in record_dict and ";" in record_dict["author"]:
                record_dict["author"] = record_dict["author"].replace("; ", " and ")

            self._table_drop_fields(record_dict=record_dict)

            for key in list(record_dict.keys()):
                value = record_dict[key]
                record_dict[key] = str(value)
                if value == "" or pd.isna(value):
                    del record_dict[key]

        def id_labeler(records: list) -> None:
            """Labeler for IDs."""
            for counter, record_dict in enumerate(records):
                if Fields.ID not in record_dict:
                    record_dict[Fields.ID] = str(counter + 1).zfill(6)

        load_operation.ensure_append_only(self.search_source.filename)

        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            id_labeler=id_labeler,
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=self.review_manager.logger,
        )

        return records

    def _load_md(self, *, load_operation: colrev.ops.load.Load) -> dict:
        load_operation.ensure_append_only(self.search_source.filename)

        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            logger=load_operation.review_manager.logger,
        )
        return records

    def _load_enl(self, *, load_operation: colrev.ops.load.Load) -> dict:
        def entrytype_setter(record_dict: dict) -> None:
            if "0" not in record_dict:
                keys_to_check = ["V", "N"]
                if any(k in record_dict for k in keys_to_check):
                    record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE
                else:
                    record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.INPROCEEDINGS
            else:
                if record_dict["0"] == "Journal Article":
                    record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE
                elif record_dict["0"] == "Inproceedings":
                    record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.INPROCEEDINGS
                else:
                    record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.MISC

        def field_mapper(record_dict: dict) -> None:

            key_maps = {
                ENTRYTYPES.ARTICLE: {
                    "T": Fields.TITLE,
                    "A": Fields.AUTHOR,
                    "D": Fields.YEAR,
                    "B": Fields.JOURNAL,
                    "V": Fields.VOLUME,
                    "N": Fields.NUMBER,
                    "P": Fields.PAGES,
                    "X": Fields.ABSTRACT,
                    "U": Fields.URL,
                    "8": "date",
                    "0": "type",
                },
                ENTRYTYPES.INPROCEEDINGS: {
                    "T": Fields.TITLE,
                    "A": Fields.AUTHOR,
                    "D": Fields.YEAR,
                    "B": Fields.JOURNAL,
                    "V": Fields.VOLUME,
                    "N": Fields.NUMBER,
                    "P": Fields.PAGES,
                    "X": Fields.ABSTRACT,
                    "U": Fields.URL,
                    "8": "date",
                    "0": "type",
                },
            }

            key_map = key_maps[record_dict[Fields.ENTRYTYPE]]
            for ris_key in list(record_dict.keys()):
                if ris_key in key_map:
                    standard_key = key_map[ris_key]
                    record_dict[standard_key] = record_dict.pop(ris_key)

            if Fields.AUTHOR in record_dict and isinstance(
                record_dict[Fields.AUTHOR], list
            ):
                record_dict[Fields.AUTHOR] = " and ".join(record_dict[Fields.AUTHOR])
            if Fields.EDITOR in record_dict and isinstance(
                record_dict[Fields.EDITOR], list
            ):
                record_dict[Fields.EDITOR] = " and ".join(record_dict[Fields.EDITOR])
            if Fields.KEYWORDS in record_dict and isinstance(
                record_dict[Fields.KEYWORDS], list
            ):
                record_dict[Fields.KEYWORDS] = ", ".join(record_dict[Fields.KEYWORDS])

            record_dict.pop("type", None)

            for key, value in record_dict.items():
                record_dict[key] = str(value)

        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            unique_id_field="INCREMENTAL",
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=self.review_manager.logger,
        )

        return records

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if not self.search_source.filename.is_file():
            return {}

        self._rename_erroneous_extensions()

        __load_methods = {
            ".ris": self._load_ris,
            ".bib": self._load_bib,
            ".csv": self._load_table,
            ".xls": self._load_table,
            ".xlsx": self._load_table,
            ".md": self._load_md,
            ".enl": self._load_enl,
        }

        if self.search_source.filename.suffix not in __load_methods:
            raise NotImplementedError

        records = __load_methods[self.search_source.filename.suffix](
            load_operation=load_operation
        )
        for record_id in records:
            records[record_id] = {
                k: v
                for k, v in records[record_id].items()
                if k not in FieldSet.PROVENANCE_KEYS + [Fields.SCREENING_CRITERIA]
            }
        for record in records.values():
            for key in list(record.keys()):
                if key not in FieldSet.STANDARDIZED_FIELD_KEYS:
                    record[f"colrev.unknonwn_source.{key}"] = record.pop(key)

        return records

    def _heuristically_fix_entrytypes(
        self, *, record: colrev.record.record_prep.PrepRecord
    ) -> None:
        """Prepare the record by heuristically correcting erroneous ENTRYTYPEs"""

        # Journal articles should not have booktitles/series set.
        if record.data[Fields.ENTRYTYPE] == "article":
            if Fields.BOOKTITLE in record.data and Fields.JOURNAL not in record.data:
                record.update_field(
                    key=Fields.JOURNAL,
                    value=record.data[Fields.BOOKTITLE],
                    source="unkown_source_prep",
                )
                record.remove_field(key=Fields.BOOKTITLE)
            if Fields.SERIES in record.data and Fields.JOURNAL not in record.data:
                record.update_field(
                    key=Fields.JOURNAL,
                    value=record.data[Fields.SERIES],
                    source="unkown_source_prep",
                )
                record.remove_field(key=Fields.SERIES)

        if self.search_source.filename.suffix == ".md":
            if (
                record.data[Fields.ENTRYTYPE] == "misc"
                and Fields.PUBLISHER in record.data
            ):
                record.update_field(
                    key=Fields.ENTRYTYPE, value="book", source="unkown_source_prep"
                )
            if record.data.get(Fields.YEAR, Fields.YEAR) == record.data.get(
                "date", "date"
            ):
                record.remove_field(key="date")
            if (
                "inbook" == record.data[Fields.ENTRYTYPE]
                and Fields.CHAPTER not in record.data
                and Fields.TITLE in record.data
            ):
                record.rename_field(key=Fields.TITLE, new_key=Fields.CHAPTER)

        if (
            "dissertation" in record.data.get(Fields.FULLTEXT, "NA").lower()
            and record.data[Fields.ENTRYTYPE] != "phdthesis"
        ):
            prior_e_type = record.data[Fields.ENTRYTYPE]
            record.update_field(
                key=Fields.ENTRYTYPE, value="phdthesis", source="unkown_source_prep"
            )
            self.review_manager.report_logger.info(
                f" {record.data[Fields.ID]}".ljust(self._padding, " ")
                + f"Set from {prior_e_type} to phdthesis "
                '("dissertation" in fulltext link)'
            )

        if (
            "thesis" in record.data.get(Fields.FULLTEXT, "NA").lower()
            and record.data[Fields.ENTRYTYPE] != "phdthesis"
        ):
            prior_e_type = record.data[Fields.ENTRYTYPE]
            record.update_field(
                key=Fields.ENTRYTYPE, value="phdthesis", source="unkown_source_prep"
            )
            self.review_manager.report_logger.info(
                f" {record.data[Fields.ID]}".ljust(self._padding, " ")
                + f"Set from {prior_e_type} to phdthesis "
                '("thesis" in fulltext link)'
            )

        if (
            "this thesis" in record.data.get(Fields.ABSTRACT, "NA").lower()
            and record.data[Fields.ENTRYTYPE] != "phdthesis"
        ):
            prior_e_type = record.data[Fields.ENTRYTYPE]
            record.update_field(
                key=Fields.ENTRYTYPE, value="phdthesis", source="unkown_source_prep"
            )
            self.review_manager.report_logger.info(
                f" {record.data[Fields.ID]}".ljust(self._padding, " ")
                + f"Set from {prior_e_type} to phdthesis "
                '("thesis" in abstract)'
            )

    def _format_inproceedings(
        self, *, record: colrev.record.record_prep.PrepRecord
    ) -> None:
        if (
            record.data.get(Fields.BOOKTITLE, FieldValues.UNKNOWN)
            == FieldValues.UNKNOWN
        ):
            return

        if (
            FieldValues.UNKNOWN != record.data[Fields.BOOKTITLE]
            and "inbook" != record.data[Fields.ENTRYTYPE]
        ):
            # pylint: disable=colrev-missed-constant-usage
            record.format_if_mostly_upper(Fields.BOOKTITLE, case="title")

            stripped_btitle = re.sub(r"\d{4}", "", record.data[Fields.BOOKTITLE])
            stripped_btitle = re.sub(r"\d{1,2}th", "", stripped_btitle)
            stripped_btitle = re.sub(r"\d{1,2}nd", "", stripped_btitle)
            stripped_btitle = re.sub(r"\d{1,2}rd", "", stripped_btitle)
            stripped_btitle = re.sub(r"\d{1,2}st", "", stripped_btitle)
            stripped_btitle = re.sub(r"\([A-Z]{3,6}\)", "", stripped_btitle)
            stripped_btitle = stripped_btitle.replace("Proceedings of the", "").replace(
                "Proceedings", ""
            )
            stripped_btitle = stripped_btitle.lstrip().rstrip()
            record.update_field(
                key=Fields.BOOKTITLE,
                value=stripped_btitle,
                source="unkown_source_prep",
                keep_source_if_equal=True,
            )

    def _format_article(self, record: colrev.record.record_prep.PrepRecord) -> None:
        if (
            record.data.get(Fields.JOURNAL, FieldValues.UNKNOWN) != FieldValues.UNKNOWN
            and len(record.data[Fields.JOURNAL]) > 10
            and FieldValues.UNKNOWN != record.data[Fields.JOURNAL]
        ):
            # pylint: disable=colrev-missed-constant-usage
            record.format_if_mostly_upper(Fields.JOURNAL, case="title")

        if record.data.get(Fields.VOLUME, FieldValues.UNKNOWN) != FieldValues.UNKNOWN:
            record.update_field(
                key=Fields.VOLUME,
                value=record.data[Fields.VOLUME].replace("Volume ", ""),
                source="unkown_source_prep",
                keep_source_if_equal=True,
            )

    def _format_fields(self, *, record: colrev.record.record_prep.PrepRecord) -> None:
        """Format fields"""

        if record.data.get(Fields.ENTRYTYPE, "") == "inproceedings":
            self._format_inproceedings(record=record)
        elif record.data.get(Fields.ENTRYTYPE, "") == "article":
            self._format_article(record=record)

        if record.data.get(Fields.AUTHOR, FieldValues.UNKNOWN) != FieldValues.UNKNOWN:
            # fix name format
            if (1 == len(record.data[Fields.AUTHOR].split(" ")[0])) or (
                ", " not in record.data[Fields.AUTHOR]
            ):
                record.update_field(
                    key=Fields.AUTHOR,
                    value=colrev.record.record_prep.PrepRecord.format_author_field(
                        record.data[Fields.AUTHOR]
                    ),
                    source="unkown_source_prep",
                    keep_source_if_equal=True,
                )
            # Replace nicknames in parentheses
            record.data[Fields.AUTHOR] = re.sub(
                r"\([^)]*\)", "", record.data[Fields.AUTHOR]
            )
            record.data[Fields.AUTHOR] = (
                record.data[Fields.AUTHOR].replace("  ", " ").rstrip()
            )

        if record.data.get(Fields.TITLE, FieldValues.UNKNOWN) != FieldValues.UNKNOWN:
            record.format_if_mostly_upper(Fields.TITLE)

        if Fields.PAGES in record.data:
            record.unify_pages_field()
            if (
                not re.match(r"^\d*$", record.data[Fields.PAGES])
                and not re.match(r"^\d*--\d*$", record.data[Fields.PAGES])
                and not re.match(r"^[xivXIV]*--[xivXIV]*$", record.data[Fields.PAGES])
            ):
                self.review_manager.report_logger.info(
                    f" {record.data[Fields.ID]}:".ljust(self._padding, " ")
                    + f"Unusual pages: {record.data[Fields.PAGES]}"
                )

        if (
            Fields.URL in record.data
            and Fields.FULLTEXT in record.data
            and record.data[Fields.URL] == record.data[Fields.FULLTEXT]
        ):
            record.remove_field(key=Fields.FULLTEXT)

        if Fields.LANGUAGE in record.data:
            try:
                self.language_service.unify_to_iso_639_3_language_codes(record=record)
                record.update_field(
                    key=Fields.LANGUAGE,
                    value=record.data[Fields.LANGUAGE],
                    source="unkown_source_prep",
                    keep_source_if_equal=True,
                )
            except colrev_exceptions.InvalidLanguageCodeException:
                del record.data[Fields.LANGUAGE]

    def _remove_redundant_fields(
        self, *, record: colrev.record.record_prep.PrepRecord
    ) -> None:
        if (
            record.data[Fields.ENTRYTYPE] == "article"
            and Fields.JOURNAL in record.data
            and Fields.BOOKTITLE in record.data
        ):
            similarity_journal_booktitle = fuzz.partial_ratio(
                record.data[Fields.JOURNAL].lower(),
                record.data[Fields.BOOKTITLE].lower(),
            )
            if similarity_journal_booktitle / 100 > 0.9:
                record.remove_field(key=Fields.BOOKTITLE)

        if record.data.get(Fields.PUBLISHER, "") in ["researchgate.net"]:
            record.remove_field(key=Fields.PUBLISHER)

        if (
            record.data[Fields.ENTRYTYPE] == "inproceedings"
            and Fields.JOURNAL in record.data
            and Fields.BOOKTITLE in record.data
        ):
            similarity_journal_booktitle = fuzz.partial_ratio(
                record.data[Fields.JOURNAL].lower(),
                record.data[Fields.BOOKTITLE].lower(),
            )
            if similarity_journal_booktitle / 100 > 0.9:
                record.remove_field(key=Fields.JOURNAL)

    def _impute_missing_fields(
        self, *, record: colrev.record.record_prep.PrepRecord
    ) -> None:
        if "date" in record.data and Fields.YEAR not in record.data:
            year = re.search(r"\d{4}", record.data["date"])
            if year:
                record.update_field(
                    key=Fields.YEAR,
                    value=year.group(0),
                    source="unkown_source_prep",
                    keep_source_if_equal=True,
                )

    def _unify_special_characters(
        self, *, record: colrev.record.record_prep.PrepRecord
    ) -> None:
        # Remove html entities
        for field in list(record.data.keys()):
            if field in [Fields.TITLE, Fields.AUTHOR, Fields.JOURNAL, Fields.BOOKTITLE]:
                record.data[field] = re.sub(r"\s+", " ", record.data[field])
                record.data[field] = re.sub(self.HTML_CLEANER, "", record.data[field])

    def prepare(
        self,
        record: colrev.record.record_prep.PrepRecord,
        source: colrev.settings.SearchSource,
    ) -> colrev.record.record.Record:
        """Source-specific preparation for unknown sources"""

        if not record.has_quality_defects() or record.masterdata_is_curated():
            return record

        # we may assign fields heuristically (e.g., to colrev.pubmed.pubmedid)

        self._heuristically_fix_entrytypes(
            record=record,
        )

        self._impute_missing_fields(record=record)

        self._format_fields(record=record)

        self._remove_redundant_fields(record=record)

        self._unify_special_characters(record=record)

        return record



#!/usr/bin/env python3
"""Scripts to run a selective backward search."""
from __future__ import annotations

from pathlib import Path

import colrev.record.record
from colrev.constants import Fields
from colrev.constants import SearchType


def main(*, search_operation: colrev.ops.search.Search, bws: str) -> None:
    """Run a selective backward search (interactively on the cli)"""

    items = search_operation.review_manager.dataset.read_next_record(
        conditions=[{Fields.ID: bws}]
    )
    record_dict = list(items)[0]
    record = colrev.record.record.Record(record_dict)

    if not record.data.get(Fields.FILE, "NA").endswith(".pdf"):
        return
    if not record.get_tei_filename().is_file():
        search_operation.review_manager.logger.debug(
            f" creating tei: {record.data['ID']}"
        )
    tei = search_operation.review_manager.get_tei(
        pdf_path=Path(record.data[Fields.FILE]),
        tei_path=record.get_tei_filename(),
    )

    search_source = colrev.settings.SearchSource(
        endpoint="colrev.unknown_source",
        filename=Path("data/search/complementary_backward_search.bib"),
        search_type=SearchType.OTHER,
        search_parameters={},
        comment="",
    )
    feed = search_source.get_api_feed(
        review_manager=search_operation.review_manager,
        source_identifier="bws_id",
        update_only=False,
    )

    # print list
    tei_recs = tei.get_references()
    for i, tei_rec_dict in enumerate(tei_recs):
        tei_rec = colrev.record.record.Record(tei_rec_dict)
        print(f"{i}  : {tei_rec.format_bib_style()}")

    # import as record
    selection = ""
    while selection == "" or selection.isdigit():
        selection = input("Select record to import")
        if selection.isdigit() and 0 <= int(selection) < len(tei_recs) + 1:
            selected_record = tei_recs[int(selection)]
            selected_record["bws_id"] = f"{bws}/#{selection}"
            print(selected_record)
            feed.add_update_record(
                retrieved_record=colrev.record.record.Record(selected_record)
            )

        feed.save()



#! /usr/bin/env python
"""Exclude collections as a prep operation"""
from __future__ import annotations

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields

# pylint: disable=duplicate-code


# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class ExcludeCollectionsPrep:
    """Prepares records by excluding collection entries (e.g., proceedings)"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    ci_supported: bool = Field(default=True)

    source_correction_hint = "check with the developer"
    always_apply_changes = True

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)

    def prepare(
        self,
        record: colrev.record.record_prep.PrepRecord,
    ) -> colrev.record.record.Record:
        """Prepare records by excluding collections (proceedings)"""

        if record.data[Fields.ENTRYTYPE].lower() == "proceedings":
            record.prescreen_exclude(reason="collection/proceedings")

        return record



#! /usr/bin/env python
"""Exclude records based on language as a prep operation"""
from __future__ import annotations

import re
import statistics

import zope.interface
from pydantic import Field

import colrev.env.language_service
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import FieldValues
from colrev.constants import RecordState


# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class ExcludeLanguagesPrep:
    """Prepares records by excluding ones that are not in the languages_to_include"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    ci_supported: bool = Field(default=True)

    source_correction_hint = "check with the developer"
    always_apply_changes = True

    def __init__(self, *, prep_operation: colrev.ops.prep.Prep, settings: dict) -> None:
        self.settings = self.settings_class(**settings)

        # Note : the following objects have heavy memory footprints and should be
        # class (not object) properties to keep parallel processing as
        # efficient as possible (the object is passed to each thread)
        languages_to_include = ["eng"]
        # if not prep_operation.review_manager.in_ci_environment():
        self.language_service = colrev.env.language_service.LanguageService()

        prescreen_package_endpoints = (
            prep_operation.review_manager.settings.prescreen.prescreen_package_endpoints
        )
        if "scope_prescreen" in [s["endpoint"] for s in prescreen_package_endpoints]:
            for scope_prescreen in [
                s
                for s in prescreen_package_endpoints
                if "scope_prescreen" == s["endpoint"]
            ]:
                languages_to_include.extend(
                    scope_prescreen.get("LanguageScope", ["eng"])
                )
        self.language_service.validate_iso_639_3_language_codes(
            lang_code_list=languages_to_include
        )
        self.languages_to_include = list(set(languages_to_include))

    def _title_has_multiple_languages(self, *, title: str) -> bool:
        if "[" not in title:
            return False
        split_titles = [
            re.sub(r"\d", "", x.rstrip().rstrip("]")) for x in title.split("[")
        ]
        min_len = statistics.mean(len(x) for x in split_titles) - 20
        if all(len(x) > min_len for x in split_titles):
            return True
        return False

    def prepare(
        self, record: colrev.record.record_prep.PrepRecord
    ) -> colrev.record.record.Record:
        """Prepare the record by excluding records whose metadata is not in English"""

        # Note : other languages are not yet supported
        # because the dedupe does not yet support cross-language merges

        if record.data.get(Fields.TITLE, FieldValues.UNKNOWN) == FieldValues.UNKNOWN:
            record.remove_field(key=Fields.LANGUAGE)
            return record

        if (
            Fields.LANGUAGE in record.data
            and record.data[Fields.LANGUAGE] in self.languages_to_include
        ):
            return record

        # To avoid misclassifications for short titles
        if len(record.data.get(Fields.TITLE, "")) < 30:
            # If language not in record, add language
            # (always - needed in dedupe.)
            record.data[Fields.LANGUAGE] = "eng"
            return record

        if not self._title_has_multiple_languages(
            title=record.data.get(Fields.TITLE, "")
        ):
            language = self.language_service.compute_language(
                text=record.data[Fields.TITLE]
            )
            # Note: classification of non-english titles is not reliable.
            # Other languages should be checked in man-prep.
            record.update_field(
                key=Fields.LANGUAGE,
                value=language,
                source="LanguageDetector",
                note="",
                append_edit=False,
            )
        else:
            # Deal with title fields containing titles in two or more languages
            split_titles = [
                x.rstrip().rstrip("]") for x in record.data[Fields.TITLE].split("[")
            ]
            for i, split_title in enumerate(split_titles):
                lang_split_title = self.language_service.compute_language(
                    text=split_title
                )
                if 0 == i:
                    record.update_field(
                        key=Fields.TITLE,
                        value=split_title.rstrip(),
                        source="LanguageDetector_split",
                    )
                    record.update_field(
                        key=Fields.LANGUAGE,
                        value=lang_split_title,
                        source="LanguageDetector_split",
                    )
                else:
                    record.update_field(
                        key=f"title_{lang_split_title}",
                        value=split_title.rstrip("]"),
                        source="LanguageDetector_split",
                    )

        if record.data.get(Fields.LANGUAGE, "") == "":
            record.update_field(
                key=Fields.TITLE,
                value=record.data.get(Fields.TITLE, ""),
                source="LanguageDetector",
                note="language-not-found",
            )
            record.set_status(RecordState.md_needs_manual_preparation)
            return record

        if record.data.get(Fields.LANGUAGE, "") not in self.languages_to_include:
            record.prescreen_exclude(
                reason=f"language of title not in [{','.join(self.languages_to_include)}]"
            )
        return record



#! /usr/bin/env python
"""SearchSource: EBSCOHost"""
from __future__ import annotations

import re
from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import FieldValues
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class EbscoHostSearchSource:
    """EBSCOHost"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings

    endpoint = "colrev.ebsco_host"
    # https://connect.ebsco.com/s/article/
    # What-is-the-Accession-Number-AN-in-EBSCOhost-records?language=en_US
    # Note : ID is the accession number.
    source_identifier = "{{ID}}"
    search_types = [SearchType.DB]

    ci_supported: bool = Field(default=False)
    heuristic_status = SearchSourceHeuristicStatus.supported

    db_url = "https://search.ebscohost.com/"

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.search_source = self.settings_class(**settings)
        self.review_manager = source_operation.review_manager
        self.source_operation = source_operation

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for EBSCOHost"""

        result = {"confidence": 0.0}

        if data.count("@") >= 1:
            if "URL = {https://search.ebscohost.com/" in data:
                if re.match(r"@.*{\d{17}\,\n", data):
                    result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint"""

        search_source = operation.create_db_source(
            search_source_cls=cls,
            params={},
        )
        operation.add_source_and_search(search_source)
        return search_source

    def search(self, rerun: bool) -> None:
        """Run a search of EbscoHost"""

        if self.search_source.search_type == SearchType.DB:
            self.source_operation.run_db_search(  # type: ignore
                search_source_cls=self.__class__,
                source=self.search_source,
            )
        else:
            raise NotImplementedError

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def _load_bib(self) -> dict:
        def field_mapper(record_dict: dict) -> None:
            for key in list(record_dict.keys()):
                if key not in ["ID", "ENTRYTYPE"]:
                    record_dict[key.lower()] = record_dict.pop(key)

        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            logger=self.review_manager.logger,
            unique_id_field="ID",
            field_mapper=field_mapper,
        )
        return records

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            return self._load_bib()

        raise NotImplementedError

    def prepare(
        self,
        record: colrev.record.record_prep.PrepRecord,
        source: colrev.settings.SearchSource,
    ) -> colrev.record.record.Record:
        """Source-specific preparation for EBSCOHost"""

        record.format_if_mostly_upper(Fields.AUTHOR, case=Fields.TITLE)
        record.format_if_mostly_upper(Fields.TITLE, case=Fields.TITLE)

        if record.data.get(Fields.PAGES) == "N.PAG -- N.PAG":
            record.data[Fields.PAGES] = FieldValues.UNKNOWN
        return record



#! /usr/bin/env python
"""SearchSource: Pubmed"""
from __future__ import annotations

import typing
from multiprocessing import Lock
from pathlib import Path
from urllib.parse import urlparse

import pandas as pd
import requests
import zope.interface
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
import colrev.record.record_prep
import colrev.record.record_similarity
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import RecordState
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType
from colrev.packages.pubmed.src import pubmed_api

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class PubMedSearchSource:
    """Pubmed"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    source_identifier = "pubmedid"
    search_types = [
        SearchType.DB,
        SearchType.API,
        SearchType.MD,
    ]
    endpoint = "colrev.pubmed"

    ci_supported: bool = Field(default=True)
    heuristic_status = SearchSourceHeuristicStatus.supported

    db_url = "https://pubmed.ncbi.nlm.nih.gov/"
    _pubmed_md_filename = Path("data/search/md_pubmed.bib")

    def __init__(
        self,
        *,
        source_operation: colrev.process.operation.Operation,
        settings: typing.Optional[dict] = None,
    ) -> None:
        self.review_manager = source_operation.review_manager
        if settings:
            # Pubmed as a search_source
            self.search_source = self.settings_class(**settings)
        else:
            # Pubmed as an md-prep source
            pubmed_md_source_l = [
                s
                for s in self.review_manager.settings.sources
                if s.filename == self._pubmed_md_filename
            ]
            if pubmed_md_source_l:
                self.search_source = pubmed_md_source_l[0]
            else:
                self.search_source = colrev.settings.SearchSource(
                    endpoint=self.endpoint,
                    filename=self._pubmed_md_filename,
                    search_type=SearchType.MD,
                    search_parameters={},
                    comment="",
                )

            self.pubmed_lock = Lock()

        self.source_operation = source_operation
        self.quality_model = self.review_manager.get_qm()
        _, self.email = self.review_manager.get_committer()

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Pubmed"""

        result = {"confidence": 0.0}

        # Simple heuristic:
        if "PMID,Title,Authors,Citation,First Author,Journal/Book," in data:
            result["confidence"] = 1.0
            return result
        if "PMID- " in data:
            result["confidence"] = 0.7
            return result

        if "pmid " in data:
            if data.count(" pmid ") > data.count("\n@"):
                result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        params_dict = {}
        if params:
            if params.startswith("http"):
                params_dict = {Fields.URL: params}
            else:
                for item in params.split(";"):
                    key, value = item.split("=")
                    params_dict[key] = value

        search_type = operation.select_search_type(
            search_types=cls.search_types, params=params_dict
        )

        if search_type == SearchType.DB:
            search_source = operation.create_db_source(
                search_source_cls=cls,
                params=params_dict,
            )

        elif search_type == SearchType.API:
            if len(params_dict) == 0:
                search_source = operation.create_api_source(endpoint=cls.endpoint)

            # pylint: disable=colrev-missed-constant-usage
            elif "url" in params_dict:
                host = urlparse(params_dict["url"]).hostname

                if host and host.endswith("pubmed.ncbi.nlm.nih.gov"):
                    query = {
                        "query": params_dict["url"].replace(
                            "https://pubmed.ncbi.nlm.nih.gov/?term=", ""
                        )
                    }

                    filename = operation.get_unique_filename(file_path_string="pubmed")
                    # params = (
                    # "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi?db=pubmed&term="
                    #     + params
                    # )
                    search_source = colrev.settings.SearchSource(
                        endpoint=cls.endpoint,
                        filename=filename,
                        search_type=SearchType.API,
                        search_parameters=query,
                        comment="",
                    )
                else:
                    raise NotImplementedError
            else:
                raise NotImplementedError

        else:
            raise NotImplementedError

        operation.add_source_and_search(search_source)
        return search_source

    def _validate_source(self) -> None:
        """Validate the SearchSource (parameters etc.)"""

        source = self.search_source
        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")

        if source.filename.name != self._pubmed_md_filename.name:
            if "query" not in source.search_parameters:
                raise colrev_exceptions.InvalidQueryException(
                    f"Source missing query search_parameter ({source.filename})"
                )

            # if "query_file" in source.search_parameters:
            # ...

        self.review_manager.logger.debug(f"SearchSource {source.filename} validated")

    def check_availability(
        self, *, source_operation: colrev.process.operation.Operation
    ) -> None:
        """Check status (availability) of the Pubmed API"""

        try:
            # pylint: disable=duplicate-code
            test_rec = {
                Fields.AUTHOR: "Nazzaro, P and Manzari, M and Merlo, M and Triggiani, R and "
                "Scarano, A and Ciancio, L and Pirrelli, A",
                Fields.TITLE: "Distinct and combined vascular effects of ACE blockade and "
                "HMG-CoA reductase inhibition in hypertensive subjects",
                Fields.ENTRYTYPE: "article",
                "pubmedid": "10024335",
            }

            api = pubmed_api.PubmedAPI(
                parameters=self.search_source.search_parameters,
                email=self.email,
                session=self.review_manager.get_cached_session(),
            )
            returned_record = api.query_id(pubmed_id=test_rec["pubmedid"])

            if returned_record:
                assert returned_record.data[Fields.TITLE] == test_rec[Fields.TITLE]
                assert returned_record.data[Fields.AUTHOR] == test_rec[Fields.AUTHOR]
            else:
                if not self.review_manager.force_mode:
                    raise colrev_exceptions.ServiceNotAvailableException("Pubmed")
        except (requests.exceptions.RequestException, IndexError, KeyError) as exc:
            print(exc)
            if not self.review_manager.force_mode:
                raise colrev_exceptions.ServiceNotAvailableException("Pubmed") from exc

    def _get_masterdata_record(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool,
        timeout: int,
    ) -> colrev.record.record.Record:
        try:
            api = pubmed_api.PubmedAPI(
                parameters=self.search_source.search_parameters,
                email=self.email,
                session=self.review_manager.get_cached_session(),
                logger=self.review_manager.logger,
            )

            retrieved_record = api.query_id(pubmed_id=record.data["pubmedid"])

            if not retrieved_record:
                raise colrev_exceptions.RecordNotFoundInPrepSourceException(
                    msg="Pubmed: no records retrieved"
                )

            if not colrev.record.record_similarity.matches(record, retrieved_record):
                return record

            try:
                self.pubmed_lock.acquire(timeout=60)

                # Note : need to reload file because the object is not shared between processes
                pubmed_feed = self.search_source.get_api_feed(
                    review_manager=self.review_manager,
                    source_identifier=self.source_identifier,
                    update_only=False,
                    prep_mode=True,
                )

                pubmed_feed.add_update_record(retrieved_record)

                record.merge(
                    retrieved_record,
                    default_source=retrieved_record.data[Fields.ORIGIN][0],
                )

                record.set_masterdata_complete(
                    source=retrieved_record.data[Fields.ORIGIN][0],
                    masterdata_repository=self.review_manager.settings.is_curated_repo(),
                )
                record.set_status(RecordState.md_prepared)
                if save_feed:
                    pubmed_feed.save()
                try:
                    self.pubmed_lock.release()
                except ValueError:
                    pass

                return record
            except (colrev_exceptions.NotFeedIdentifiableException,):
                try:
                    self.pubmed_lock.release()
                except ValueError:
                    pass

                return record

        except (
            requests.exceptions.RequestException,
            OSError,
            IndexError,
            colrev_exceptions.RecordNotFoundInPrepSourceException,
            colrev_exceptions.RecordNotParsableException,
        ):
            pass
        return record

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Retrieve masterdata from Pubmed based on similarity with the record provided"""

        # To test the metadata provided for a particular pubmed-id use:
        # https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&id=10075143&rettype=xml&retmode=text

        if (
            len(record.data.get(Fields.TITLE, "")) < 35
            and Fields.PUBMED_ID not in record.data
        ):
            return record

        # at this point, we coujld validate metadata
        # if "pubmedid" not in record.data:
        #    record = self._check_doi_masterdata(record=record)

        # remove the following if we match basd on similarity
        if Fields.PUBMED_ID not in record.data:
            return record

        record = self._get_masterdata_record(
            prep_operation=prep_operation,
            record=record,
            timeout=timeout,
            save_feed=save_feed,
        )

        return record

    def _run_api_search(
        self,
        *,
        pubmed_feed: colrev.ops.search_api_feed.SearchAPIFeed,
        rerun: bool,
    ) -> None:
        if rerun:
            self.review_manager.logger.info(
                "Performing a search of the full history (may take time)"
            )

        api = pubmed_api.PubmedAPI(
            parameters=self.search_source.search_parameters,
            email=self.email,
            session=self.review_manager.get_cached_session(),
            logger=self.review_manager.logger,
        )

        for record in api.get_query_return():
            try:
                # Note : discard "empty" records
                if "" == record.data.get(Fields.AUTHOR, "") and "" == record.data.get(
                    Fields.TITLE, ""
                ):
                    self.review_manager.logger.warning(f"Skipped record: {record.data}")
                    continue
                prep_record = colrev.record.record_prep.PrepRecord(record.data)

                if Fields.D_PROV in prep_record.data:
                    del prep_record.data[Fields.D_PROV]

                added = pubmed_feed.add_update_record(prep_record)

                # Note : only retrieve/update the latest deposits (unless in rerun mode)
                if not added and not rerun:
                    # problem: some publishers don't necessarily
                    # deposit papers chronologically
                    break
            except colrev_exceptions.NotFeedIdentifiableException:
                print("Cannot set id for record")
                continue

        pubmed_feed.save()

    def _run_md_search(
        self,
        *,
        pubmed_feed: colrev.ops.search_api_feed.SearchAPIFeed,
    ) -> None:

        api = pubmed_api.PubmedAPI(
            parameters=self.search_source.search_parameters,
            email=self.email,
            session=self.review_manager.get_cached_session(),
            logger=self.review_manager.logger,
        )

        for feed_record_dict in pubmed_feed.feed_records.values():
            feed_record = colrev.record.record.Record(feed_record_dict)

            try:
                retrieved_record = api.query_id(pubmed_id=feed_record_dict["pubmedid"])

                if retrieved_record.data["pubmedid"] == feed_record.data["pubmedid"]:
                    pubmed_feed.add_update_record(retrieved_record)

            except (
                colrev_exceptions.RecordNotFoundInPrepSourceException,
                colrev_exceptions.NotFeedIdentifiableException,
                colrev_exceptions.SearchSourceException,
            ):
                continue

        pubmed_feed.save()

    def search(self, rerun: bool) -> None:
        """Run a search of Pubmed"""

        self._validate_source()

        pubmed_feed = self.search_source.get_api_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        if self.search_source.search_type == SearchType.MD:
            self._run_md_search(pubmed_feed=pubmed_feed)

        elif self.search_source.search_type == SearchType.API:
            self._run_api_search(
                pubmed_feed=pubmed_feed,
                rerun=rerun,
            )

        elif self.search_source.search_type == SearchType.DB:
            self.source_operation.run_db_search(  # type: ignore
                search_source_cls=self.__class__,
                source=self.search_source,
            )
            return
        else:
            raise NotImplementedError

    def _load_csv(self) -> dict:
        def entrytype_setter(record_dict: dict) -> None:
            record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE

        def field_mapper(record_dict: dict) -> None:
            record_dict[Fields.TITLE] = record_dict.pop("Title", "")
            record_dict[Fields.JOURNAL] = record_dict.pop("Journal/Book", "")
            record_dict[Fields.YEAR] = record_dict.pop("Publication Year", "")
            record_dict[Fields.URL] = record_dict.pop("URL", "")
            record_dict[Fields.DOI] = record_dict.pop("DOI", "")
            record_dict[f"{self.endpoint}.nihms_id"] = record_dict.pop("NIHMS ID", "")
            record_dict[Fields.PUBMED_ID] = record_dict.pop("PMID", "")
            record_dict[Fields.PMCID] = record_dict.pop("PMCID", "")
            record_dict[f"{self.endpoint}.create_date"] = record_dict.pop(
                "Create Date", ""
            )

            author_list = record_dict.pop("Authors", "").split(", ")
            for i, author_part in enumerate(author_list):
                author_field_parts = author_part.split(" ")
                author_list[i] = (
                    author_field_parts[0] + ", " + " ".join(author_field_parts[1:])
                )
            record_dict[Fields.AUTHOR] = " and ".join(author_list)

            if "Citation" in record_dict:
                details_part = record_dict["Citation"]
                details_part = details_part[details_part.find(";") + 1 :]
                details_part = details_part[: details_part.find(".")]
                if ":" in details_part:
                    record_dict[Fields.PAGES] = details_part[
                        details_part.find(":") + 1 :
                    ]
                    details_part = details_part[: details_part.find(":")]
                if "(" in details_part:
                    record_dict[Fields.NUMBER] = details_part[
                        details_part.find("(") + 1 : -1
                    ]
                    details_part = details_part[: details_part.find("(")]
                record_dict[Fields.VOLUME] = details_part

            record_dict.pop("First Author", None)
            record_dict.pop("Citation", None)

            for key in list(record_dict.keys()):
                value = record_dict[key]
                record_dict[key] = str(value)
                if value == "" or pd.isna(value):
                    del record_dict[key]

        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            unique_id_field="PMID",
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=self.review_manager.logger,
        )
        return records

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".csv":
            return self._load_csv()

        if self.search_source.filename.suffix == ".bib":
            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                logger=self.review_manager.logger,
            )
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for Pubmed"""

        if "colrev.pubmed.first_author" in record.data:
            record.remove_field(key="colrev.pubmed.first_author")

        if Fields.AUTHOR in record.data:
            record.data[Fields.AUTHOR] = (
                colrev.record.record_prep.PrepRecord.format_author_field(
                    record.data[Fields.AUTHOR]
                )
            )

        return record



#! /usr/bin/env python
"""Consolidation of metadata based on DOI metadata as a prep operation"""
from __future__ import annotations

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.packages.doi_org.src.doi_org as doi_connector
import colrev.record.record
from colrev.constants import Fields


# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class DOIMetadataPrep:
    """Prepares records based on doi.org metadata"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=True)

    source_correction_hint = (
        "ask the publisher to correct the metadata"
        + " (see https://www.crossref.org/blog/"
        + "metadata-corrections-updates-and-additions-in-metadata-manager/"
    )
    always_apply_changes = False

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.prep_operation = prep_operation
        self.review_manager = prep_operation.review_manager

    def prepare(
        self, record: colrev.record.record_prep.PrepRecord
    ) -> colrev.record.record.Record:
        """Prepare the record by retrieving its metadata from doi.org"""

        if Fields.DOI not in record.data:
            return record
        doi_connector.DOIConnector.retrieve_doi_metadata(
            review_manager=self.review_manager,
            record=record,
            timeout=self.prep_operation.timeout,
        )
        doi_connector.DOIConnector.get_link_from_doi(
            record=record,
            review_manager=self.review_manager,
        )
        return record



#! /usr/bin/env python
"""SearchSource: Wiley"""
from __future__ import annotations

from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class WileyOnlineLibrarySearchSource:
    """Wiley"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    endpoint = "colrev.wiley"
    source_identifier = "url"
    search_types = [SearchType.DB]

    ci_supported: bool = Field(default=False)
    heuristic_status = SearchSourceHeuristicStatus.supported

    db_url = "https://onlinelibrary.wiley.com/"

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.search_source = self.settings_class(**settings)
        self.source_operation = source_operation
        self.review_manager = source_operation.review_manager

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Wiley"""

        result = {"confidence": 0.0}

        # Simple heuristic:
        if "eprint = {https://onlinelibrary.wiley.com/doi/pdf/" in data:
            result["confidence"] = 0.7
            return result

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        search_source = operation.create_db_source(
            search_source_cls=cls,
            params={},
        )
        operation.add_source_and_search(search_source)
        return search_source

    def search(self, rerun: bool) -> None:
        """Run a search of Wiley"""

        if self.search_source.search_type == SearchType.DB:
            self.source_operation.run_db_search(  # type: ignore
                search_source_cls=self.__class__,
                source=self.search_source,
            )
            return

        raise NotImplementedError

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                logger=self.review_manager.logger,
                unique_id_field="ID",
            )
            for record_dict in records.values():
                if "eprint" not in record_dict:
                    continue
                record_dict[Fields.FULLTEXT] = record_dict.pop("eprint")
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for Wiley"""

        if record.data.get(Fields.ENTRYTYPE, "") == "inbook":
            record.rename_field(key=Fields.TITLE, new_key=Fields.CHAPTER)
            record.rename_field(key=Fields.BOOKTITLE, new_key=Fields.TITLE)

        return record



# pylint: disable=missing-module-docstring
try:
    from importlib.metadata import version
except ImportError:  # pragma: no cover
    # For Python < 3.8
    from importlib_metadata import version  # type: ignore

__version__ = version("colrev")



#! /usr/bin/env python
"""Template for a custom SearchSource PackageEndpoint"""
from __future__ import annotations

from pathlib import Path

import zope.interface

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_settings
import colrev.process.operation
import colrev.record.record
from colrev.constants import Fields


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class CustomSearch:
    """Class for custom search scripts"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    source_identifier = "custom"

    def __init__(
        self,
        *,
        source_operation: colrev.ops.search.Search,
        settings: dict,
    ) -> None:
        self.search_source: colrev.settings.SearchSource = self.settings_class(
            **settings
        )
        self.review_manager = source_operation.review_manager

    def search(self, rerun: bool) -> None:
        """Run the search"""

        feed = self.search_source.get_api_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )
        retrieved_record = {
            Fields.ID: "ID00001",
            Fields.ENTRYTYPE: "article",
            Fields.AUTHOR: "Smith, M.",
            Fields.TITLE: "Editorial",
            Fields.JOURNAL: "nature",
            Fields.YEAR: "2020",
        }

        feed.add_update_record(colrev.record.record.Record(retrieved_record))
        feed.save()

    @classmethod
    def validate_search_params(cls, query: str) -> None:
        """Validate the search parameters"""

        if " SCOPE " not in query:
            raise colrev_exceptions.InvalidQueryException(
                "CROSSREF queries require a SCOPE section"
            )

        scope = query[query.find(" SCOPE ") :]
        if "journal_issn" not in scope:
            raise colrev_exceptions.InvalidQueryException(
                "CROSSREF queries require a journal_issn field in the SCOPE section"
            )

    def heuristic(
        self, filename: Path, data: str  # pylint: disable=unused-argument
    ) -> dict:
        """Heuristic to identify the custom source"""

        result = {"confidence": 0}

        return result

    def load(
        self,
        load_operation: colrev.ops.load.Load,  # pylint: disable=unused-argument
    ) -> dict:
        """Load fixes for the custom source"""
        records = {"ID1": {"ID": "ID1", "title": "..."}}

        return records

    def prepare(
        self, record: colrev.record.record.Record
    ) -> colrev.record.record.Record:
        """Source-specific preparation for the custom source"""

        return record



#! /usr/bin/env python
"""SearchSource: Unpaywall"""
from __future__ import annotations

import typing
from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType
from colrev.packages.unpaywall.src.api import UnpaywallAPI

# pylint: disable=unused-argument


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class UnpaywallSearchSource:
    """Unpaywall Search Source"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings

    source_identifier = "doi"
    search_types = [SearchType.API]
    endpoint = "colrev.unpaywall"

    ci_supported: bool = Field(default=False)
    heuristic_status = SearchSourceHeuristicStatus.oni
    docs_link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/packages/unpaywall/README.md"
    )

    def __init__(
        self,
        *,
        source_operation: colrev.process.operation.Operation,
        settings: typing.Optional[dict] = None,
    ) -> None:
        self.review_manager = source_operation.review_manager
        if settings:
            # Unpaywall as a search_source
            self.search_source = self.settings_class(**settings)
        else:
            self.search_source = colrev.settings.SearchSource(
                endpoint=self.endpoint,
                filename=Path("data/search/unpaywall.bib"),
                search_type=SearchType.API,
                search_parameters={},
                comment="",
            )

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Unpaywall"""
        result = {"confidence": 0.0}
        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""

        params_dict = {}
        if params:
            if params.startswith("http"):
                params_dict = {Fields.URL: params}
            else:
                for item in params.split(";"):
                    key, value = item.split("=")
                    params_dict[key] = value

        if len(params_dict) == 0:
            search_source = operation.create_api_source(endpoint=cls.endpoint)

        # pylint: disable=colrev-missed-constant-usage
        elif "https://api.unpaywall.org/v2/search" in params_dict["url"]:
            query = (
                params_dict["url"]
                .replace("https://api.unpaywall.org/v2/search?", "")
                .replace("https://api.unpaywall.org/v2/search/?", "")
                .lstrip("&")
            )

            parameter_pairs = query.split("&")
            search_parameters = {}
            for parameter in parameter_pairs:
                key, value = parameter.split("=")
                search_parameters[key] = value

            filename = operation.get_unique_filename(file_path_string="unpaywall")

            search_parameters["query"] = (
                UnpaywallAPI.decode_html_url_encoding_to_string(
                    query=search_parameters["query"]
                )
            )

            search_source = colrev.settings.SearchSource(
                endpoint=cls.endpoint,
                filename=filename,
                search_type=SearchType.API,
                search_parameters=search_parameters,
                comment="",
            )
        else:
            raise colrev_exceptions.PackageParameterError(
                f"Cannot add UNPAYWALL endpoint with query {params}"
            )

        operation.add_source_and_search(search_source)

    def _run_api_search(
        self, *, unpaywall_feed: colrev.ops.search_api_feed.SearchAPIFeed, rerun: bool
    ) -> None:

        api = UnpaywallAPI(self.search_source.search_parameters)
        for record in api.get_query_records():
            unpaywall_feed.add_update_record(record)

        unpaywall_feed.save()

    def search(self, rerun: bool) -> None:
        """Run a search of Unpaywall"""

        unpaywall_feed = self.search_source.get_api_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        if self.search_source.search_type == SearchType.API:
            self._run_api_search(unpaywall_feed=unpaywall_feed, rerun=rerun)
        else:
            raise NotImplementedError

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                logger=self.review_manager.logger,
            )
            return records

        raise NotImplementedError

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for Unpaywall"""
        return record



#! /usr/bin/env python
"""CLI interface for manual preparation of PDFs"""
from __future__ import annotations

import os
import platform
import re
import subprocess
import textwrap
from pathlib import Path

import inquirer
import zope.interface
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Colors
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import RecordState

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PDFPrepManInterface)
class CoLRevCLIPDFManPrep:
    """Manually prepare PDFs based on a CLI (not yet implemented)"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=False)

    _to_skip: int = 0

    def __init__(
        self,
        *,
        pdf_prep_man_operation: colrev.ops.pdf_prep_man.PDFPrepMan,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.review_manager = pdf_prep_man_operation.review_manager
        self.pdf_prep_man_operation = pdf_prep_man_operation

    def _update_metadata(
        self, *, record: colrev.record.record.Record
    ) -> colrev.record.record.Record:
        questions = [
            inquirer.List(
                "field",
                message="Update metadata fields:",
                choices=[
                    "Author",
                    "Container title",
                    "Title",
                    "Volume",
                    "Number",
                    "Pages",
                    "Save",
                ],
            ),
        ]
        while True:
            answers = inquirer.prompt(questions)
            user_selection = answers["field"]

            if user_selection == "Save":
                break
            if user_selection == "Author":
                author = input("Authors:")
                record.update_field(
                    key=Fields.AUTHOR, value=author, source="manual_correction"
                )
            elif user_selection == "Container title":
                if Fields.JOURNAL in record.data:
                    journal = input("Journal:")
                    record.update_field(
                        key=Fields.JOURNAL, value=journal, source="manual_correction"
                    )
                if Fields.BOOKTITLE in record.data:
                    booktitle = input("Booktitle:")
                    record.update_field(
                        key=Fields.BOOKTITLE,
                        value=booktitle,
                        source="manual_correction",
                    )
            elif user_selection == "Title":
                title = input("Title:")
                record.update_field(
                    key=Fields.TITLE, value=title, source="manual_correction"
                )
            elif user_selection == "Volume":
                volume = input("Volume:")
                record.update_field(
                    key=Fields.VOLUME, value=volume, source="manual_correction"
                )
            elif user_selection == "Number":
                number = input("Number:")
                record.update_field(
                    key=Fields.NUMBER, value=number, source="manual_correction"
                )
            elif user_selection == "Pages":
                pages = input("Pages:")
                record.update_field(
                    key=Fields.PAGES, value=pages, source="manual_correction"
                )
            user_selection = ""

        return record

    def _open_pdf(self, *, filepath: Path) -> None:
        try:
            system_platform = platform.system().lower()

            if os.getenv("CODESPACES") is not None:
                subprocess.run(["code", filepath], check=True)
            elif system_platform == "darwin":  # macOS
                subprocess.run(["open", filepath], check=True)
            elif system_platform == "windows":
                subprocess.run(["start", "", filepath], check=True, shell=True)
            elif system_platform == "linux":
                subprocess.run(["xdg-open", filepath], check=True)
            else:
                print("Unsupported operating system.")
        except Exception as exc:  # pylint: disable=broad-exception-caught
            print(f"Error: {exc}")

    def _remove_page(
        self,
        *,
        user_selection: str,
        filepath: Path,
    ) -> None:
        if user_selection == "Remove coverpage":
            try:
                self.pdf_prep_man_operation.extract_coverpage(filepath=filepath)
            except colrev_exceptions.InvalidPDFException:
                pass
        elif user_selection == "Remove last page":
            try:
                self.pdf_prep_man_operation.extract_lastpage(filepath=filepath)
            except colrev_exceptions.InvalidPDFException:
                pass
        elif user_selection == "Remove page range":
            range_str_questions = [
                inquirer.Text(
                    "range_str",
                    message="Page range to remove (e.g., 1-3):",
                    validate=lambda _, x: re.match(r"(\d)+-(\d)+", x),
                )
            ]
            answers = inquirer.prompt(range_str_questions)
            range_str = answers["range_str"]
            pages_to_exclude = list(
                range(
                    int(range_str[: range_str.find("-")]) - 1,
                    int(range_str[range_str.find("-") + 1 :]),
                )
            )
            try:
                self.pdf_prep_man_operation.extract_pages(
                    filepath=filepath, pages_to_remove=pages_to_exclude
                )
            except colrev_exceptions.InvalidPDFException:
                pass

    def _is_inside_wsl(self) -> bool:
        return "wsl" in platform.uname().release.lower()

    def _man_pdf_prep_item(
        self,
        *,
        filepath: Path,
        record: colrev.record.record.Record,
    ) -> None:
        if not self._is_inside_wsl():
            self._open_pdf(filepath=filepath)

        # if PDF > 100 pages, we may check on which page we find the title & print

        questions = [
            inquirer.List(
                "prep_decision",
                message="Prepared?",
                choices=[
                    "Skip",
                    "Yes",
                    "No (delete)",
                    "Remove coverpage",
                    "Remove last page",
                    "Remove page range",
                    "Metadata needs to be updated",
                    "Quit",
                ],
            ),
        ]
        while True:
            answers = inquirer.prompt(questions)
            user_selection = answers["prep_decision"]

            # if user_selection.startswith("s"):
            #     if user_selection[1:].isdigit():
            #         self._to_skip = int(user_selection[1:])
            #     return
            if user_selection == "Skip":
                return
            if user_selection in [
                "Remove coverpage",
                "Remove last page",
                "Remove page range",
            ]:
                self._remove_page(
                    user_selection=user_selection,
                    filepath=filepath,
                )
            elif user_selection == "Yes":
                self.pdf_prep_man_operation.set_pdf_man_prepared(record)
                return
            elif user_selection == "No (delete)":
                record.remove_field(key=Fields.FILE)
                record.remove_field(key=Fields.PDF_ID)
                record.set_status(RecordState.pdf_needs_manual_retrieval)
                if filepath.is_file():
                    filepath.unlink()
                return
            elif user_selection == "Metadata needs to be updated":
                self._update_metadata(record=record)

            elif user_selection == "Quit":
                raise QuitPressedException()

    def _print_pdf_prep_man(self, record: colrev.record.record.Record) -> None:
        """Print the record for pdf-prep-man operations"""
        # pylint: disable=too-many-branches
        ret_str = ""
        if Fields.FILE in record.data:
            ret_str += (
                f"\nfile: {Colors.ORANGE}{record.data[Fields.FILE]}{Colors.END}\n\n"
            )

        pdf_prep_note = record.get_field_provenance(key=Fields.FILE)

        if "author_not_in_first_pages" in pdf_prep_note["note"]:
            ret_str += f"{Colors.RED}{record.data.get(Fields.AUTHOR, 'no-author')}{Colors.END}\n"
        else:
            ret_str += f"{Colors.GREEN}{record.data.get(Fields.AUTHOR, 'no-author')}{Colors.END}\n"

        if "title_not_in_first_pages" in pdf_prep_note["note"]:
            ret_str += (
                f"{Colors.RED}{record.data.get(Fields.TITLE, 'no title')}{Colors.END}\n"
            )
        else:
            ret_str += f"{Colors.GREEN}{record.data.get(Fields.TITLE, 'no title')}{Colors.END}\n"

        if record.data[Fields.ENTRYTYPE] == ENTRYTYPES.ARTICLE:
            ret_str += (
                f"{record.data.get(Fields.JOURNAL, 'no-journal')} "
                f"({record.data.get(Fields.YEAR, 'no-year')}) "
                f"{record.data.get(Fields.VOLUME, 'no-volume')}"
                f"({record.data.get(Fields.NUMBER, '')})"
            )
            if Fields.PAGES in record.data:
                if "nr_pages_not_matching" in pdf_prep_note["note"]:
                    ret_str += (
                        f", {Colors.RED}pp.{record.data[Fields.PAGES]}{Colors.END}\n"
                    )
                else:
                    ret_str += (
                        f", pp.{Colors.GREEN}{record.data[Fields.PAGES]}{Colors.END}\n"
                    )
            else:
                ret_str += "\n"
        elif record.data[Fields.ENTRYTYPE] == ENTRYTYPES.INPROCEEDINGS:
            ret_str += f"{record.data.get(Fields.BOOKTITLE, 'no-booktitle')}\n"
        if Fields.ABSTRACT in record.data:
            lines = textwrap.wrap(
                record.data[Fields.ABSTRACT], 100, break_long_words=False
            )
            ret_str += f"\nAbstract: {lines.pop(0)}\n"
            ret_str += "\n".join(lines) + "\n"

        if Fields.URL in record.data:
            ret_str += f"\nurl: {record.data[Fields.URL]}\n"

        print(ret_str)

    def _man_pdf_prep_item_init(
        self,
        *,
        records: dict,
        item: dict,
        stat: str,
    ) -> dict:
        current_platform = platform.system()
        if current_platform in ["Linux", "Darwin"]:
            os.system("clear")
        else:
            os.system("cls")

        # to do : if authors mismatch: color those that do/do not match
        print(stat)
        record = colrev.record.record.Record(item)
        file_provenance = record.get_field_provenance(key=Fields.FILE)

        self._print_pdf_prep_man(record)
        record_dict = records[item[Fields.ID]]
        record = colrev.record.record.Record(record_dict)
        if RecordState.pdf_needs_manual_preparation != record_dict[Fields.STATUS]:
            return record_dict

        print(
            "Manual preparation needed:"
            f" {Colors.RED}{file_provenance['note']}{Colors.END}"
        )

        filepath = self.review_manager.path / Path(record_dict[Fields.FILE])
        if not filepath.is_file():
            filepath = self.review_manager.paths.pdf / f"{record_dict['ID']}.pdf"
        if not filepath.is_file():
            input(
                f"{Colors.ORANGE}Warning: PDF file for record {record_dict['ID']} not found. "
                f"Manual retrieval may be required.{Colors.END}"
            )
            return records

        try:
            record.data.update(colrev_pdf_id=record.get_colrev_pdf_id(filepath))
        except colrev_exceptions.InvalidPDFException:
            pass

        if filepath.is_file():
            self._man_pdf_prep_item(
                filepath=filepath,
                record=record,
            )

        else:
            print(f"File does not exist ({record.data[Fields.ID]})")

        self.review_manager.dataset.save_records_dict(records)

        return records

    def pdf_prep_man(self, records: dict) -> dict:
        """Prepare PDF manually based on a cli"""

        self.review_manager.logger.info("Loading data for pdf_prep_man")
        pdf_prep_man_data = self.pdf_prep_man_operation.get_data()
        records = self.review_manager.dataset.load_records_dict()

        for i, item in enumerate(pdf_prep_man_data["items"]):
            if self._to_skip > 0:
                self._to_skip -= 1
                continue
            try:
                stat = str(i + 1) + "/" + str(pdf_prep_man_data["nr_tasks"])
                records = self._man_pdf_prep_item_init(
                    records=records,
                    item=item,
                    stat=stat,
                )
            except QuitPressedException:
                break

        self.review_manager.dataset.save_records_dict(records)

        if self.pdf_prep_man_operation.pdfs_prepared_manually():
            if input("Create commit (y/n)?") == "y":
                self.review_manager.dataset.create_commit(
                    msg="Prepare PDFs manually",
                    manual_author=True,
                )
        else:
            self.review_manager.logger.info(
                "Prepare PDFs manually. Afterwards, use colrev pdf-get-man"
            )

        return records


class QuitPressedException(Exception):
    """Quit-pressed exception"""



#! /usr/bin/env python
"""Traces records and changes through history."""
from __future__ import annotations

import time
import typing

import dictdiffer

import colrev.process.operation
from colrev.constants import Colors
from colrev.constants import OperationsType

if typing.TYPE_CHECKING:  # pragma: no cover
    import git.objects.commit


class Trace(colrev.process.operation.Operation):
    """Trace a record through history"""

    type = OperationsType.check

    def __init__(self, *, review_manager: colrev.review_manager.ReviewManager) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
        )

    def _print_diff(self, *, diff: dict, color: str, lpad: int = 5) -> None:
        formatted_diff = self.review_manager.p_printer.pformat(diff)
        lines = formatted_diff.splitlines()
        print(
            color
            + "\n".join(["".join([" " * lpad]) + line for line in lines])
            + Colors.END
        )

    def _print_record_changes(
        self,
        *,
        commit: git.objects.commit.Commit,
        records_dict: dict,
        record_id: str,
        prev_record: dict,
    ) -> dict:
        record = records_dict[record_id]

        diffs = list(dictdiffer.diff(prev_record, record))

        if len(diffs) > 0:
            if not self.review_manager.verbose_mode:  # pragma: no cover
                commit_message_first_line = str(commit.message).partition("\n")[0]
                print(
                    "\n\n"
                    + time.strftime(
                        "%Y-%m-%d %H:%M",
                        time.gmtime(commit.committed_date),
                    )
                    + f" {commit} ".ljust(40, " ")
                    + f" {commit_message_first_line} (by {commit.author.name})"
                )

            for diff in diffs:
                if diff[0] == "add":
                    self._print_diff(diff=diff, color=Colors.GREEN)
                if diff[0] == "change":
                    self._print_diff(diff=diff, color=Colors.ORANGE)
                if diff[0] == "delete":  # pragma: no cover
                    self._print_diff(diff=diff, color=Colors.RED)

        prev_record = record
        return prev_record

    @colrev.process.operation.Operation.decorate()
    def main(self, *, record_id: str) -> None:
        """Trace a record (main entrypoint)"""

        self.review_manager.logger.info(f"Trace record by ID: {record_id}")
        # Ensure the path uses forward slashes, which is compatible with Git's path handling

        revlist = self.review_manager.dataset.get_repo().iter_commits(
            paths=self.review_manager.paths.RECORDS_FILE_GIT
        )

        prev_record: dict = {}
        for commit in reversed(list(revlist)):
            filecontents = (
                commit.tree / self.review_manager.paths.RECORDS_FILE_GIT
            ).data_stream.read()
            commit_message_first_line = str(commit.message).partition("\n")[0]

            if self.review_manager.verbose_mode:
                print(
                    "\n\n"
                    + time.strftime(
                        "%Y-%m-%d %H:%M",
                        time.gmtime(commit.committed_date),
                    )
                    + f" {commit} ".ljust(40, " ")
                    + f" {commit_message_first_line} (by {commit.author.name})"
                )

            records_dict = colrev.loader.load_utils.loads(
                load_string=filecontents.decode("utf-8"),
                implementation="bib",
                logger=self.review_manager.logger,
            )

            if record_id not in records_dict:
                if self.review_manager.verbose_mode:
                    print(f"record {record_id} not in commit.")
                continue

            prev_record = self._print_record_changes(
                commit=commit,
                records_dict=records_dict,
                record_id=record_id,
                prev_record=prev_record,
            )



#! /usr/bin/env python
"""Removal of broken URLs (error 500) a prep operation"""
from __future__ import annotations

import requests
import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields


# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class RemoveError500URLsPrep:
    """Prepares records by removing urls that are not available"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    ci_supported: bool = Field(default=True)

    source_correction_hint = "check with the developer"
    always_apply_changes = True

    requests_headers = {
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) "
        "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36"
    }

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.prep_operation = prep_operation
        self.review_manager = prep_operation.review_manager

    def prepare(
        self, record: colrev.record.record_prep.PrepRecord
    ) -> colrev.record.record.Record:
        """Prepare the record by removing URLs with 500 errors"""

        session = self.review_manager.get_cached_session()

        try:
            if Fields.URL in record.data:
                ret = session.request(
                    "GET",
                    record.data[Fields.URL],
                    headers=self.requests_headers,
                    timeout=60,
                )
                if ret.status_code >= 500:
                    record.remove_field(key=Fields.URL)
        except requests.exceptions.RequestException:
            pass
        try:
            if Fields.FULLTEXT in record.data:
                ret = session.request(
                    "GET",
                    record.data[Fields.FULLTEXT],
                    headers=self.requests_headers,
                    timeout=self.prep_operation.timeout,
                )
                if ret.status_code >= 500:
                    record.remove_field(key=Fields.FULLTEXT)
        except requests.exceptions.RequestException:
            pass

        return record



#! /usr/bin/env python
"""Discovering and using packages."""
from __future__ import annotations

import colrev.process.operation
from colrev.constants import Colors
from colrev.constants import EndpointType
from colrev.constants import OperationsType


def _get_endpoint_with_type(operation: colrev.process.operation.Operation) -> tuple:
    settings = operation.review_manager.settings
    package_type_dict = {
        OperationsType.search: {
            "package_type": EndpointType.search_source,
            "endpoint_location": settings.sources,
        },
        OperationsType.prep: {
            "package_type": EndpointType.prep,
            "endpoint_location": settings.prep.prep_rounds[0].prep_package_endpoints,
        },
        OperationsType.prep_man: {
            "package_type": EndpointType.prep_man,
            "endpoint_location": settings.prep.prep_man_package_endpoints,
        },
        OperationsType.dedupe: {
            "package_type": EndpointType.dedupe,
            "endpoint_location": settings.dedupe.dedupe_package_endpoints,
        },
        OperationsType.prescreen: {
            "package_type": EndpointType.prescreen,
            "endpoint_location": settings.prescreen.prescreen_package_endpoints,
        },
        OperationsType.pdf_get: {
            "package_type": EndpointType.pdf_get,
            "endpoint_location": settings.pdf_get.pdf_get_package_endpoints,
        },
        OperationsType.pdf_get_man: {
            "package_type": EndpointType.pdf_get_man,
            "endpoint_location": settings.pdf_get.pdf_get_man_package_endpoints,
        },
        OperationsType.pdf_prep: {
            "package_type": EndpointType.pdf_prep,
            "endpoint_location": settings.pdf_prep.pdf_prep_package_endpoints,
        },
        OperationsType.pdf_prep_man: {
            "package_type": EndpointType.pdf_prep_man,
            "endpoint_location": settings.pdf_prep.pdf_prep_man_package_endpoints,
        },
        OperationsType.screen: {
            "package_type": EndpointType.screen,
            "endpoint_location": settings.screen.screen_package_endpoints,
        },
        OperationsType.data: {
            "package_type": EndpointType.data,
            "endpoint_location": settings.data.data_package_endpoints,
        },
    }
    endpoints_in_settings = package_type_dict[operation.type]["endpoint_location"]
    package_type = package_type_dict[operation.type]["package_type"]
    return endpoints_in_settings, package_type


def add_package_to_settings(
    package_manager: colrev.package_manager.package_manager.PackageManager,
    *,
    operation: colrev.process.operation.Operation,
    package_identifier: str,
    params: str,
) -> None:
    """Add a package_endpoint (for cli usage)"""

    operation.review_manager.logger.info(
        f"{Colors.GREEN}Add {operation.type} "
        f"package:{Colors.END} {package_identifier}"
    )

    endpoints_in_settings, package_type = _get_endpoint_with_type(operation)
    e_class = package_manager.get_package_endpoint_class(
        package_type=package_type,
        package_identifier=package_identifier,
    )

    if hasattr(e_class, "add_endpoint"):
        e_class.add_endpoint(operation=operation, params=params)  # type: ignore

    else:
        add_package = {"endpoint": package_identifier}
        endpoints_in_settings.append(add_package)  # type: ignore
        operation.review_manager.save_settings()
        operation.review_manager.dataset.create_commit(
            msg=f"Add {operation.type} {package_identifier}",
        )
